# Copyright 2017 Nokia
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from vitrage_tempest_plugin.tests.e2e.test_basic_actions import TestActionsBase

LOG = logging.getLogger(__name__)


class BaseTestEvents(TestActionsBase):
    """Test class for Vitrage event API"""

    def _check_alarms(self):
        api_alarms = self.vitrage_client.alarm.list(vitrage_id='all',
                                                    all_tenants=True)
        if api_alarms:
            return True, api_alarms
        return False, api_alarms

    def _post_event(self, event_time, event_type, details):
        event_time_iso = event_time.isoformat()
        self.vitrage_client.event.post(event_time_iso, event_type, details)



# Copyright 2017 ZTE, Nokia
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_serialization import jsonutils
from testtools import matchers

from vitrage_tempest_plugin.tests.base import BaseVitrageTempest
from vitrage_tempest_plugin.tests.base import IsEmpty
from vitrage_tempest_plugin.tests.base import IsNotEmpty
from vitrage_tempest_plugin.tests.common.constants import \
    CINDER_VOLUME_DATASOURCE
from vitrage_tempest_plugin.tests.common.constants import \
    NOVA_INSTANCE_DATASOURCE
from vitrage_tempest_plugin.tests.common.constants import VertexProperties as \
    VProps
from vitrage_tempest_plugin.tests.common import neutron_utils
from vitrage_tempest_plugin.tests.common import nova_utils
from vitrage_tempest_plugin.tests import utils
from vitrageclient.exceptions import ClientException

LOG = logging.getLogger(__name__)


class TestResource(BaseVitrageTempest):
    """Test class for Vitrage resource API tests."""

    properties = (VProps.VITRAGE_ID,
                  VProps.VITRAGE_TYPE,
                  VProps.ID,
                  VProps.STATE,
                  VProps.VITRAGE_AGGREGATED_STATE)

    @classmethod
    def setUpClass(cls):
        super(TestResource, cls).setUpClass()
        cls.instances = nova_utils.create_instances(num_instances=1,
                                                    set_public_network=True)

    # noinspection PyPep8Naming
    @classmethod
    def tearDownClass(cls):
        super(TestResource, cls).tearDownClass()
        nova_utils.delete_created_instances(cls.instances)

    @utils.tempest_logger
    def test_compare_cli_vs_api_resource_list(self):
        """resource list """
        api_resources = self.vitrage_client.resource.list(
            all_tenants=True)

        LOG.info("api_resources = %s", api_resources)

        cli_resources = utils.run_vitrage_command(
            'vitrage resource list --all -f json')

        self._compare_resources(api_resources, cli_resources)

    @utils.tempest_logger
    def test_default_resource_list(self):
        """resource list with default query

        get the resources: network, instance, port, zone, host, cluster
        """
        # TODO(e0ne): split this test to verify that only network,
        # instance and port are returned to non-admin user.
        resources = self.vitrage_client.resource.list(all_tenants=False)
        has_shared_network = neutron_utils.get_neutron_network('shared')
        has_private_network = neutron_utils.get_neutron_network('private')
        num_resources = 6
        if has_shared_network:
            num_resources += 1
        if has_private_network:
            num_resources += 1
        self.assertThat(resources, matchers.HasLength(num_resources))

    @utils.tempest_logger
    def test_resource_list_with_all_tenants(self):
        """resource list with all tenants

        get the resources:

        """
        instances = None
        try:
            resources_before = self.vitrage_client.resource.list(
                all_tenants=True)
            instances = nova_utils.create_instances(num_instances=1,
                                                    set_public_network=True)
            resources = self.vitrage_client.resource.list(all_tenants=True)

            self.assertEqual(len(resources_before) + 2, len(resources))

        finally:
            if instances:
                nova_utils.delete_created_instances(instances)

    @utils.tempest_logger
    def test_resource_list_with_existing_type(self):
        """resource list with existing type

        get the resource: one instance
        """
        resources = self.vitrage_client.resource.list(
            resource_type=NOVA_INSTANCE_DATASOURCE,
            all_tenants=True)
        self.assertThat(resources, matchers.HasLength(1))

    @utils.tempest_logger
    def test_resource_list_with_no_existing_type(self):
        """resource list with no existing type"""
        resources = self.vitrage_client.resource.list(
            resource_type=CINDER_VOLUME_DATASOURCE,
            all_tenants=True)
        self.assertThat(resources, IsEmpty())

    @utils.tempest_logger
    def test_resource_list_with_query_existing(self):
        resources = self.vitrage_client.resource.list(
            resource_type=NOVA_INSTANCE_DATASOURCE,
            all_tenants=True,
            query='{"==": {"name": "vm-0"}}'
        )
        self.assertThat(resources, matchers.HasLength(1))

    @utils.tempest_logger
    def test_resource_list_with_query_none_existing(self):
        resources = self.vitrage_client.resource.list(
            resource_type=NOVA_INSTANCE_DATASOURCE,
            all_tenants=True,
            query='{"==": {"name": "kuku-does-not-exist"}}'
        )
        self.assertThat(resources, matchers.HasLength(0))

    @utils.tempest_logger
    def test_resource_count(self):
        resources = self.vitrage_client.resource.count(
            resource_type=NOVA_INSTANCE_DATASOURCE,
            all_tenants=True,
            query='{"==": {"name": "vm-0"}}'
        )
        self.assertThat(resources, matchers.HasLength(1))
        self.assertEqual(1, resources[NOVA_INSTANCE_DATASOURCE])

    @utils.tempest_logger
    def test_resource_count_empty(self):
        resources = self.vitrage_client.resource.count(
            resource_type=NOVA_INSTANCE_DATASOURCE,
            all_tenants=True,
            query='{"==": {"name": "kuku-does-not-exist"}}'
        )
        self.assertThat(resources, matchers.HasLength(0))

    def test_resource_show_with_no_existing_resource(self):
        """resource_show test no existing resource"""
        try:

            self.assertRaises(ClientException,
                              self.vitrage_client.resource.get,
                              'test_for_no_existing')
        finally:
            nova_utils.delete_all_instances()

    def _compare_resources(self, api_resources, cli_resources):
        self.assertThat(api_resources, IsNotEmpty(),
                        'The resources taken from rest api is empty')
        self.assertThat(cli_resources, IsNotEmpty(),
                        'The resources taken from rest api is empty')

        sorted_cli_resources = sorted(
            jsonutils.loads(cli_resources),
            key=lambda resource: resource["ID"])
        sorted_api_resources = sorted(
            api_resources,
            key=lambda resource: resource["vitrage_id"])

        self.assertEqual(len(sorted_cli_resources),
                         len(sorted_api_resources), 'cli = %s --> api = %s' %
                         (sorted_cli_resources, sorted_api_resources))

        for cli_resource, api_resource in \
                zip(sorted_cli_resources, sorted_api_resources):

            self.assertEqual(
                cli_resource.get("ID").lower(),
                api_resource.get(VProps.VITRAGE_ID).lower())
            self.assertEqual(
                cli_resource.get("Type").lower(),
                api_resource.get(VProps.VITRAGE_TYPE).lower())
            self.assertEqual(
                cli_resource.get("Data Source ID").lower(),
                api_resource.get(VProps.ID).lower())

            # Don't compare the state for Neutron ports. See bug #1776921
            if 'neutron.port' != cli_resource.get("Type").lower():
                self.assertEqual(
                    cli_resource.get("State").lower(),
                    api_resource.get(VProps.VITRAGE_OPERATIONAL_STATE).lower())



# Copyright 2017 Nokia
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from datetime import datetime
from oslo_log import log as logging
from testtools import matchers

from vitrage_tempest_plugin.tests.api.event.base import BaseTestEvents
from vitrage_tempest_plugin.tests.base import IsEmpty
from vitrage_tempest_plugin.tests.common.constants import EntityCategory
from vitrage_tempest_plugin.tests.common.constants import EventProperties \
    as EventProps
from vitrage_tempest_plugin.tests.common.constants import VertexProperties \
    as VProps
from vitrage_tempest_plugin.tests.common.vitrage_utils import DOWN
from vitrage_tempest_plugin.tests.utils import wait_for_answer

LOG = logging.getLogger(__name__)


class TestEvents(BaseTestEvents):
    """Test class for Vitrage event API"""

    def test_send_doctor_event_without_resource_id(self):
        """Sending an event in Doctor format should result in an alarm"""
        self._test_send_doctor_event(
            self._create_doctor_event_details('host123', DOWN))

    def test_send_doctor_event_without_resource_id_v2(self):
        """Sending an event in Doctor format should result in an alarm"""
        self._test_send_doctor_event(
            self._create_doctor_event_details('host457', DOWN))

    def _test_send_doctor_event(self, details):
        try:
            # post an event to the message bus
            event_time = datetime.now()
            event_type = 'compute.host.down'
            self._post_event(event_time, event_type, details)
            api_alarms = wait_for_answer(2, 0.5, self._check_alarms)

            # expect to get a 'host down alarm', generated by Doctor datasource
            self.assertThat(api_alarms, matchers.HasLength(1),
                            'Expected host down alarm')

            alarm = api_alarms[0]
            event_time_tz = str(event_time.strftime('%Y-%m-%dT%H:%M:%SZ'))
            self._check_alarm(alarm, event_time_tz, event_type, details)

            event_time = datetime.now()
            details['status'] = 'up'
            self._post_event(event_time, event_type, details)

            api_alarms = wait_for_answer(2, 0.5, self._check_alarms)
            self.assertIsNotNone(api_alarms)
            self.assertThat(api_alarms, IsEmpty(), 'Expected no alarms on '
                                                   'host')

        except Exception as e:
            LOG.exception(e)
            raise

    def _check_alarm(self, alarm, event_time, event_type, details):
        self.assertEqual(EntityCategory.ALARM, alarm[VProps.VITRAGE_CATEGORY])
        self.assertEqual(event_type, alarm[VProps.NAME])
        self.assertEqual(event_time, alarm[EventProps.TIME])
        self.assertEqual(details['status'], alarm['status'])
        self.assertFalse(alarm[VProps.VITRAGE_IS_DELETED])
        self.assertFalse(alarm[VProps.VITRAGE_IS_PLACEHOLDER])

    @staticmethod
    def _create_doctor_event_details(hostname, status):
        return {
            'hostname': hostname,
            'source': 'sample_monitor',
            'cause': 'another alarm',
            'severity': 'critical',
            'status': status,
            'monitor_id': 'sample monitor',
            'monitor_event_id': '456',
        }



# Copyright 2016 Nokia
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from functools import wraps

import time

from oslo_log import log as logging
from tempest.common import credentials_factory as common_creds
from tempest import config

import subprocess

CONF = config.CONF
LOG = logging.getLogger(__name__)

TIMESTAMP_FORMAT = '%Y-%m-%dT%H:%M:%SZ'


def run_vitrage_command(command):
    admin_creds = common_creds.get_configured_admin_credentials()

    auth_url = CONF.identity.uri_v3
    auth_url_param = '--os-auth-url ' + auth_url

    # USERNAME
    user = admin_creds.username
    user_param = '--os-user-name ' + user

    # PASSWORD
    password = admin_creds.password
    password_param = '--os-password ' + password

    # PROJECT_NAME
    project_name = admin_creds.project_name
    project_name_param = '--os-project-name ' + project_name

    # USER_DOMAIN_ID
    user_domain_id = admin_creds.user_domain_id
    user_domain_id_param = '--os-user-domain-id ' + user_domain_id

    # PROJECT_DOMAIN_ID
    project_domain_id = admin_creds.project_domain_id
    project_domain_id_par = '--os-project-domain-id ' + project_domain_id

    full_command = '%s %s %s %s %s %s %s' % (command, user_param,
                                             password_param,
                                             project_name_param,
                                             auth_url_param,
                                             user_domain_id_param,
                                             project_domain_id_par)

    LOG.debug('Full command: %s', full_command)

    child = subprocess.Popen(full_command,
                             shell=True,
                             executable="/bin/bash",
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)

    stdout, stderr = child.communicate()
    if stderr:
        LOG.error('error from command %(command)s = %(error)s',
                  {'error': stderr, 'command': full_command})

    output = stdout.decode('utf-8')

    LOG.debug('cli stdout: %s', output)

    if child.returncode:
        LOG.error('process return code is not 0 : return code = %d',
                  child.returncode)
    return output


def uni2str(text):
    return text.encode('ascii', 'ignore')


def tempest_logger(func):
    func_name = func.__name__

    @wraps(func)
    def func_name_print_func(*args, **kwargs):
        LOG.info('Test Start: ' + func_name)
        result = func(*args, **kwargs)
        LOG.info('Test End: ' + func_name)
        return result

    return func_name_print_func


def wait_for_answer(max_waiting, time_between_attempts, func, **kwargs):
    """time_between_attempts should be in range of 0 to 1"""
    status, res = False, None
    start_time = time.time()
    while time.time() - start_time < max_waiting:
        time.sleep(time_between_attempts)
        status, res = func(**kwargs)
        if status:
            return res
    LOG.info("wait for answer- False")
    return res


def wait_for_status(max_waiting, func, **kwargs):
    count = 0
    while count < max_waiting:
        if func(**kwargs):
            return True
        count += 1
        time.sleep(2)
    LOG.error("wait_for_status - False")
    return False


def filter_data(data, filter_, match_filter=True):
    if match_filter:
        return {k for k in data if k in filter_}
    return {k for k in data if k not in filter_}



# Copyright 2015 - Alcatel-Lucent
# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

AODH_DATASOURCE = 'aodh'
CINDER_VOLUME_DATASOURCE = 'cinder.volume'
HEAT_STACK_DATASOURCE = 'heat.stack'
NEUTRON_NETWORK_DATASOURCE = 'neutron.network'
NEUTRON_PORT_DATASOURCE = 'neutron.port'
NOVA_HOST_DATASOURCE = 'nova.host'
NOVA_INSTANCE_DATASOURCE = 'nova.instance'
NOVA_ZONE_DATASOURCE = 'nova.zone'
OPENSTACK_CLUSTER = 'openstack.cluster'
VITRAGE_DATASOURCE = 'vitrage'

DOCTOR_DATASOURCE = 'doctor'


class VertexProperties(object):
    VITRAGE_CATEGORY = 'vitrage_category'
    VITRAGE_TYPE = 'vitrage_type'
    VITRAGE_ID = 'vitrage_id'
    VITRAGE_STATE = 'vitrage_state'
    VITRAGE_IS_DELETED = 'vitrage_is_deleted'
    VITRAGE_IS_PLACEHOLDER = 'vitrage_is_placeholder'
    VITRAGE_SAMPLE_TIMESTAMP = 'vitrage_sample_timestamp'
    VITRAGE_AGGREGATED_STATE = 'vitrage_aggregated_state'
    VITRAGE_OPERATIONAL_STATE = 'vitrage_operational_state'
    VITRAGE_AGGREGATED_SEVERITY = 'vitrage_aggregated_severity'
    VITRAGE_OPERATIONAL_SEVERITY = 'vitrage_operational_severity'
    VITRAGE_RESOURCE_ID = 'vitrage_resource_id'
    ID = 'id'
    STATE = 'state'
    PROJECT_ID = 'project_id'
    UPDATE_TIMESTAMP = 'update_timestamp'
    NAME = 'name'
    SEVERITY = 'severity'
    IS_MARKED_DOWN = 'is_marked_down'
    INFO = 'info'
    GRAPH_INDEX = 'graph_index'
    RAWTEXT = 'rawtext'
    RESOURCE_ID = 'resource_id'
    RESOURCE_NAME = 'resource_name'
    VITRAGE_RESOURCE_TYPE = 'vitrage_resource_type'
    RESOURCE = 'resource'
    IS_REAL_VITRAGE_ID = 'is_real_vitrage_id'


class EdgeProperties(object):
    RELATIONSHIP_TYPE = 'relationship_type'
    VITRAGE_IS_DELETED = 'vitrage_is_deleted'
    UPDATE_TIMESTAMP = 'update_timestamp'


class EdgeLabel(object):
    """Define *some* edge labels

    Note that edge labels are not restricted to the values in this class, and
    other datasources can defined their own edge labels.
    """
    ON = 'on'
    CONTAINS = 'contains'
    CAUSES = 'causes'
    ATTACHED = 'attached'
    ATTACHED_PUBLIC = 'attached_public'
    ATTACHED_PRIVATE = 'attached_private'
    CONNECT = 'connect'
    MANAGED_BY = 'managed_by'
    COMPRISED = 'comprised'

    @staticmethod
    def labels():
        return [value for label, value in vars(EdgeLabel).items()
                if not label.startswith(('_', 'labels'))]


class DatasourceAction(object):
    SNAPSHOT = 'snapshot'
    INIT_SNAPSHOT = 'init_snapshot'
    UPDATE = 'update'


class UpdateMethod(object):
    NONE = 'none'
    PULL = 'pull'
    PUSH = 'push'


class EntityCategory(object):
    RESOURCE = 'RESOURCE'
    ALARM = 'ALARM'

    @staticmethod
    def categories():
        return [value for category, value in vars(EntityCategory).items()
                if not category.startswith(('_', 'categories'))]


class DatasourceProperties(object):
    ENTITY_TYPE = 'vitrage_entity_type'
    DATASOURCE_ACTION = 'vitrage_datasource_action'
    SAMPLE_DATE = 'vitrage_sample_date'
    EVENT_TYPE = 'vitrage_event_type'


class GraphAction(object):
    CREATE_ENTITY = 'create_entity'
    DELETE_ENTITY = 'delete_entity'
    UPDATE_ENTITY = 'update_entity'
    DELETE_RELATIONSHIP = 'delete_relationship'
    UPDATE_RELATIONSHIP = 'update_relationship'
    REMOVE_DELETED_ENTITY = 'remove_deleted_entity'
    END_MESSAGE = 'end_message'


class NotifierEventTypes(object):
    ACTIVATE_DEDUCED_ALARM_EVENT = 'vitrage.deduced_alarm.activate'
    DEACTIVATE_DEDUCED_ALARM_EVENT = 'vitrage.deduced_alarm.deactivate'
    ACTIVATE_ALARM_EVENT = 'vitrage.alarm.activate'
    DEACTIVATE_ALARM_EVENT = 'vitrage.alarm.deactivate'
    ACTIVATE_MARK_DOWN_EVENT = 'vitrage.mark_down.activate'
    DEACTIVATE_MARK_DOWN_EVENT = 'vitrage.mark_down.deactivate'
    EXECUTE_EXTERNAL_ACTION = 'vitrage.execute_external_action'


class TemplateTopologyFields(object):
    """yaml fields for topology definitions"""
    METADATA = 'metadata'
    DESCRIPTION = 'description'
    NAME = 'name'
    VERSION = 'version'

    DEFINITIONS = 'definitions'

    ENTITIES = 'entities'
    ENTITY = 'entity'
    TYPE = 'type'
    ID = 'id'

    RELATIONSHIPS = 'relationships'
    RELATIONSHIP = 'relationship'
    RELATIONSHIP_TYPE = 'relationship_type'
    SOURCE = 'source'
    TARGET = 'target'


class EventProperties(object):
    TIME_FORMAT = '%Y-%m-%dT%H:%M:%S.%f'
    TYPE = 'type'
    TIME = 'time'
    DETAILS = 'details'


class DatasourceOpts(object):
    TRANSFORMER = 'transformer'
    DRIVER = 'driver'
    UPDATE_METHOD = 'update_method'
    CHANGES_INTERVAL = 'changes_interval'
    CONFIG_FILE = 'config_file'


class TemplateTypes(object):
    STANDARD = 'standard'
    DEFINITION = 'definition'
    EQUIVALENCE = 'equivalence'

    @staticmethod
    def types():
        return [value for type, value in vars(TemplateTypes).items()
                if not type.startswith(('_', 'types'))]


class TemplateStatus(object):
    ACTIVE = 'ACTIVE'
    ERROR = 'ERROR'
    DELETING = 'DELETING'
    DELETED = 'DELETED'
    LOADING = 'LOADING'


class TenantProps(object):
    ALL_TENANTS = 'all_tenants'
    TENANT = 'tenant'
    IS_ADMIN = 'is_admin'


class AodhProperties(object):
    ALARM_ID = 'alarm_id'
    DESCRIPTION = 'description'
    ENABLED = 'enabled'
    EVENT = 'event'
    EVENT_RULE = 'event_rule'
    EVENT_TYPE = 'event_type'
    EVENT_RESOURCE_ID = 'traits.resource_id'
    NAME = 'name'
    STATE = 'state'
    PROJECT_ID = 'project_id'
    QUERY = 'query'
    REPEAT_ACTIONS = 'repeat_actions'
    RESOURCE_ID = 'resource_id'
    SEVERITY = 'severity'
    STATE_TIMESTAMP = 'state_timestamp'
    THRESHOLD = 'threshold'
    THRESHOLD_RULE = 'threshold_rule'
    RESOURCES_THRESHOLD = 'gnocchi_aggregation_by_resources_threshold'
    METRICS_THRESHOLD = 'gnocchi_aggregation_by_metrics_threshold'
    METRICS_THRESHOLD_RULE = 'gnocchi_aggregation_by_metrics_threshold_rule'
    TIMESTAMP = 'timestamp'
    TYPE = 'type'
    VITRAGE_ID = 'vitrage_id'
    DETAIL = 'detail'
    RULE = 'rule'
    RESOURCES_THRESHOLD_RULE = \
        'gnocchi_aggregation_by_resources_threshold_rule'
    USER_ID = 'user_id'
    STATE_REASON = 'state_reason'
    ID = 'id'


class OperationalAlarmSeverity(object):
    CRITICAL = 'CRITICAL'
    SEVERE = 'SEVERE'
    WARNING = 'WARNING'
    NA = 'N/A'
    OK = 'OK'


class OperationalResourceState(object):
    NA = 'N/A'
    OK = 'OK'
    TRANSIENT = 'TRANSIENT'
    SUBOPTIMAL = 'SUBOPTIMAL'
    ERROR = 'ERROR'
    DELETED = 'DELETED'



# Copyright 2017 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import time

from oslo_log import log as logging

from vitrage_tempest_plugin.tests.common.constants import DOCTOR_DATASOURCE
from vitrage_tempest_plugin.tests.common.constants import EntityCategory
from vitrage_tempest_plugin.tests.common.constants import VertexProperties \
    as VProps
from vitrage_tempest_plugin.tests.common.constants import VITRAGE_DATASOURCE
from vitrage_tempest_plugin.tests.common import general_utils as g_utils
from vitrage_tempest_plugin.tests.common import nova_utils
from vitrage_tempest_plugin.tests.common.tempest_clients import TempestClients
from vitrage_tempest_plugin.tests.common import vitrage_utils as v_utils
from vitrage_tempest_plugin.tests.e2e.test_actions_base import TestActionsBase
from vitrage_tempest_plugin.tests import utils

LOG = logging.getLogger(__name__)

TRIGGER_ALARM_1 = 'e2e.test_basic_actions.trigger.alarm1'
TRIGGER_ALARM_2 = 'e2e.test_basic_actions.trigger.alarm2'
TRIGGER_ALARM_3 = 'e2e.test_basic_actions.trigger.alarm3'
TRIGGER_ALARM_4 = 'e2e.test_basic_actions.trigger.alarm4'
TRIGGER_ALARM_5 = 'e2e.test_basic_actions.trigger.alarm5'
DEDUCED = 'e2e.test_basic_actions.deduced.alarm'

TRIGGER_ALARM_1_V3 = 'e2e.test_basic_actions.trigger.alarm1.v3'
TRIGGER_ALARM_2_V3 = 'e2e.test_basic_actions.trigger.alarm2.v3'
TRIGGER_ALARM_3_V3 = 'e2e.test_basic_actions.trigger.alarm3.v3'
TRIGGER_ALARM_4_V3 = 'e2e.test_basic_actions.trigger.alarm4.v3'
TRIGGER_ALARM_5_V3 = 'e2e.test_basic_actions.trigger.alarm5.v3'
TRIGGER_ALARM_6_V3 = 'e2e.test_basic_actions.trigger.alarm6.v3'
DEDUCED_V3 = 'e2e.test_basic_actions.deduced.alarm.v3'

TRIGGER_ALARM_2_PROPS = {
    VProps.NAME: TRIGGER_ALARM_2,
    VProps.VITRAGE_CATEGORY: EntityCategory.ALARM,
    VProps.VITRAGE_TYPE: DOCTOR_DATASOURCE,
}

TRIGGER_ALARM_2_PROPS_V3 = {
    VProps.NAME: TRIGGER_ALARM_2_V3,
    VProps.VITRAGE_CATEGORY: EntityCategory.ALARM,
    VProps.VITRAGE_TYPE: DOCTOR_DATASOURCE,
}

TRIGGER_ALARM_6_PROPS_V3 = {
    VProps.NAME: TRIGGER_ALARM_6_V3,
    VProps.VITRAGE_CATEGORY: EntityCategory.ALARM,
    VProps.VITRAGE_TYPE: DOCTOR_DATASOURCE,
}

DEDUCED_PROPS = {
    VProps.NAME: DEDUCED,
    VProps.VITRAGE_CATEGORY: EntityCategory.ALARM,
    VProps.VITRAGE_TYPE: VITRAGE_DATASOURCE,
}

DEDUCED_PROPS_V3 = {
    VProps.NAME: DEDUCED_V3,
    VProps.VITRAGE_CATEGORY: EntityCategory.ALARM,
    VProps.VITRAGE_TYPE: VITRAGE_DATASOURCE,
}


class TestBasicActions(TestActionsBase):
    @classmethod
    def setUpClass(cls):
        super(TestBasicActions, cls).setUpClass()
        cls._templates = []
        cls._templates.append(
            v_utils.add_template("e2e_test_basic_actions.yaml"))
        cls._templates.append(
            v_utils.add_template("e2e_test_basic_actions_v3.yaml"))

    @classmethod
    def tearDownClass(cls):
        for t in cls._templates:
            v_utils.delete_template(t['uuid'])

    @utils.tempest_logger
    def test_action_set_state_host(self):
        self._do_test_action_set_state_host(TRIGGER_ALARM_1)

    @utils.tempest_logger
    def test_action_set_state_host_v3(self):
        self._do_test_action_set_state_host(TRIGGER_ALARM_1_V3)

    def _do_test_action_set_state_host(self, trigger_name):
        try:

            # Do
            self._trigger_do_action(trigger_name)
            curr_host = v_utils.get_first_host()
            self.assertEqual(
                'ERROR',
                curr_host.get(VProps.VITRAGE_AGGREGATED_STATE),
                'state should change after set_state action')

            # Undo
            self._trigger_undo_action(trigger_name)
            curr_host = v_utils.get_first_host()
            self.assertEqual(
                self.orig_host.get(VProps.VITRAGE_AGGREGATED_STATE),
                curr_host.get(VProps.VITRAGE_AGGREGATED_STATE),
                'state should change after undo set_state action')
        finally:
            self._trigger_undo_action(trigger_name)

    @utils.tempest_logger
    def test_action_set_state_instance(self):
        self._do_test_action_set_state_instance(TRIGGER_ALARM_3)

    @utils.tempest_logger
    def test_action_set_state_instance_v3(self):
        self._do_test_action_set_state_instance(TRIGGER_ALARM_3_V3)

    def _do_test_action_set_state_instance(self, trigger_name):

        vm_id = ""
        try:
            vm_id = nova_utils.create_instances(set_public_network=True)[0].id

            # Do
            orig_instance = v_utils.get_first_instance(id=vm_id)
            self._trigger_do_action(trigger_name)
            curr_instance = v_utils.get_first_instance(id=vm_id)
            self.assertEqual(
                'ERROR',
                curr_instance.get(VProps.VITRAGE_AGGREGATED_STATE),
                'state should change after set_state action')

            # Undo
            self._trigger_undo_action(trigger_name)
            curr_instance = v_utils.get_first_instance(id=vm_id)
            self.assertEqual(
                orig_instance.get(VProps.VITRAGE_AGGREGATED_STATE),
                curr_instance.get(VProps.VITRAGE_AGGREGATED_STATE),
                'state should change after undo set_state action')
        finally:
            self._trigger_undo_action(trigger_name)
            nova_utils.delete_all_instances(id=vm_id)

    @utils.tempest_logger
    def test_action_mark_down_host(self):
        self._do_test_action_mark_down_host(TRIGGER_ALARM_4)

    @utils.tempest_logger
    def test_action_mark_down_host_v3(self):
        self._do_test_action_mark_down_host(TRIGGER_ALARM_4_V3)

    def _do_test_action_mark_down_host(self, trigger_name):
        try:
            host_name = self.orig_host.get(VProps.NAME)

            # Do
            self._trigger_do_action(trigger_name)
            nova_service = TempestClients.nova().services.list(
                host=host_name, binary='nova-compute')[0]
            self.assertEqual("down", nova_service.state)

            # Undo
            self._trigger_undo_action(trigger_name)
            nova_service = TempestClients.nova().services.list(
                host=host_name, binary='nova-compute')[0]
            self.assertEqual("up", nova_service.state)
        finally:
            self._trigger_undo_action(trigger_name)
            # nova.host datasource may take up to snapshot_intreval to update
            time.sleep(130)

    @utils.tempest_logger
    def test_action_mark_down_instance(self):
        self._do_test_action_mark_down_instance(TRIGGER_ALARM_5)

    @utils.tempest_logger
    def test_action_mark_down_instance_v3(self):
        self._do_test_action_mark_down_instance(TRIGGER_ALARM_5_V3)

    def _do_test_action_mark_down_instance(self, trigger_name):
        vm_id = ""
        try:
            vm_id = nova_utils.create_instances(set_public_network=True)[0].id
            # Do
            self._trigger_do_action(trigger_name)
            nova_instance = TempestClients.nova().servers.get(vm_id)
            self.assertEqual("ERROR", nova_instance.status)

            # Undo
            self._trigger_undo_action(trigger_name)
            nova_instance = TempestClients.nova().servers.get(vm_id)
            self.assertEqual("ACTIVE", nova_instance.status)
        finally:
            self._trigger_undo_action(trigger_name)
            nova_utils.delete_all_instances(id=vm_id)

    @utils.tempest_logger
    def test_action_deduce_alarm(self):
        self._do_test_action_deduce_alarm(TRIGGER_ALARM_2, DEDUCED_PROPS)

    @utils.tempest_logger
    def test_action_deduce_alarm_v3(self):
        self._do_test_action_deduce_alarm(TRIGGER_ALARM_2_V3, DEDUCED_PROPS_V3)

    def _do_test_action_deduce_alarm(self, trigger_name, deduced_props):
        try:
            host_id = self.orig_host.get(VProps.VITRAGE_ID)

            # Do
            self._trigger_do_action(trigger_name)
            self._check_deduced(1, deduced_props, host_id)

            # Undo
            self._trigger_undo_action(trigger_name)
            self._check_deduced(0, deduced_props, host_id)
        finally:
            self._trigger_undo_action(trigger_name)

    @utils.tempest_logger
    def test_action_add_causal_relationship(self):
        self._do_test_action_add_causal_relationship(TRIGGER_ALARM_2,
                                                     DEDUCED_PROPS,
                                                     TRIGGER_ALARM_2_PROPS)

    @utils.tempest_logger
    def test_action_add_causal_relationship_v3(self):
        self._do_test_action_add_causal_relationship(TRIGGER_ALARM_2_V3,
                                                     DEDUCED_PROPS_V3,
                                                     TRIGGER_ALARM_2_PROPS_V3)

    @utils.tempest_logger
    def test_action_add_raise_alarm_with_causing_alarm(self):
        self._do_test_action_add_causal_relationship(TRIGGER_ALARM_6_V3,
                                                     DEDUCED_PROPS_V3,
                                                     TRIGGER_ALARM_6_PROPS_V3)

    def _do_test_action_add_causal_relationship(self,
                                                trigger_name,
                                                deduced_props,
                                                trigger_alarm_props):
        try:
            # Do
            self._trigger_do_action(trigger_name)
            alarms = self.vitrage_client.alarm.list(
                vitrage_id=self.orig_host.get(VProps.VITRAGE_ID),
                all_tenants=True)
            self.assertTrue(len(alarms) >= 2, 'alarms %s' % alarms)

            deduced = g_utils.first_match(alarms, **deduced_props)
            trigger = g_utils.first_match(alarms, **trigger_alarm_props)

            # Get Rca for the deduced
            rca = self.vitrage_client.rca.get(deduced[VProps.VITRAGE_ID],
                                              all_tenants=True)
            self._check_rca(rca, [deduced, trigger], deduced_props)

            # Get Rca for the trigger
            rca = self.vitrage_client.rca.get(trigger[VProps.VITRAGE_ID],
                                              all_tenants=True)
            self._check_rca(rca, [deduced, trigger], trigger_alarm_props)
        finally:
            self._trigger_undo_action(trigger_name)



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log as logging
from oslo_serialization import jsonutils
from testtools import matchers

from vitrage_tempest_plugin.tests.base import BaseVitrageTempest
from vitrage_tempest_plugin.tests.base import IsNotEmpty
from vitrage_tempest_plugin.tests.common import aodh_utils
from vitrage_tempest_plugin.tests.common.constants import AODH_DATASOURCE
from vitrage_tempest_plugin.tests.common.constants import NOVA_HOST_DATASOURCE
from vitrage_tempest_plugin.tests.common.constants import \
    NOVA_INSTANCE_DATASOURCE
from vitrage_tempest_plugin.tests.common.constants \
    import OperationalAlarmSeverity
from vitrage_tempest_plugin.tests.common.constants \
    import OperationalResourceState
from vitrage_tempest_plugin.tests.common.constants import VertexProperties \
    as VProps
from vitrage_tempest_plugin.tests.common.constants import VITRAGE_DATASOURCE
from vitrage_tempest_plugin.tests.common import general_utils as g_utils
from vitrage_tempest_plugin.tests.common import nova_utils
from vitrage_tempest_plugin.tests.common import vitrage_utils

LOG = logging.getLogger(__name__)
RCA_ALARM_NAME = 'rca_test_host_alarm'
VITRAGE_ALARM_NAME = 'instance_deduce'


class BaseRcaTest(BaseVitrageTempest):

    @staticmethod
    def _clean_all():
        nova_utils.delete_all_instances()
        aodh_utils.delete_all_aodh_alarms()

    def _create_alarm(self, resource_id, alarm_name, unic=False):
        aodh_utils.create_aodh_alarm(resource_id=resource_id,
                                     name=alarm_name,
                                     unic=unic)

        list_alarms = self.vitrage_client.alarm.list(vitrage_id='all',
                                                     all_tenants=True)
        expected_alarm = g_utils.all_matches(
            list_alarms,
            resource_id=resource_id,
            vitrage_type=AODH_DATASOURCE)
        if not expected_alarm:
            return None
        return expected_alarm[0]

    def _compare_rca(self, api_rca, cli_rca):
        self.assertThat(api_rca, IsNotEmpty(),
                        'The rca taken from api is empty')
        self.assertIsNotNone(cli_rca, 'The rca taken from cli is empty')

        LOG.debug("The rca taken from cli is : %s", cli_rca)
        LOG.debug("The rca taken by api is : %s", api_rca)

        parsed_rca = jsonutils.loads(cli_rca)
        sorted_cli_graph = self._clean_timestamps(sorted(parsed_rca.items()))
        sorted_api_graph = self._clean_timestamps(sorted(api_rca.items()))
        self.assert_list_equal(sorted_cli_graph, sorted_api_graph)

    def _validate_rca(self, rca):
        self.assertThat(rca, IsNotEmpty, 'The rca is empty')
        LOG.debug("The rca alarms list is : %s", rca)

        resource_alarm = g_utils.all_matches(
            rca,
            vitrage_type=AODH_DATASOURCE,
            name=RCA_ALARM_NAME)

        deduce_alarms = g_utils.all_matches(
            rca,
            vitrage_type=VITRAGE_DATASOURCE,
            name=VITRAGE_ALARM_NAME,
            severity=OperationalAlarmSeverity.WARNING)

        self.assertThat(rca, matchers.HasLength(3))
        self.assertThat(resource_alarm, matchers.HasLength(1))
        self.assertThat(deduce_alarms, matchers.HasLength(2))

    def _validate_deduce_alarms(self, alarms, instances):
        """Validate alarm existence """
        self.assertThat(alarms, IsNotEmpty(), 'The alarms list is empty')
        LOG.debug("The alarms list is : %s", alarms)

        # Find the vitrage_id of the deduced alarms using their original id.
        vitrage_resources = self.vitrage_client.resource.list(
            all_tenants=False)
        vitrage_instance_0_id = g_utils.first_match(vitrage_resources,
                                                    id=instances[0].id)

        vitrage_instance_1_id = g_utils.first_match(vitrage_resources,
                                                    id=instances[1].id)

        # Find the deduced alarms based on their properties
        deduce_alarms_1 = g_utils.all_matches(
            alarms,
            vitrage_type=VITRAGE_DATASOURCE,
            name=VITRAGE_ALARM_NAME,
            vitrage_resource_type=NOVA_INSTANCE_DATASOURCE,
            vitrage_resource_id=vitrage_instance_0_id[VProps.VITRAGE_ID])

        deduce_alarms_2 = g_utils.all_matches(
            alarms,
            vitrage_type=VITRAGE_DATASOURCE,
            name=VITRAGE_ALARM_NAME,
            vitrage_resource_type=NOVA_INSTANCE_DATASOURCE,
            vitrage_resource_id=vitrage_instance_1_id[VProps.VITRAGE_ID])

        self.assertThat(alarms, matchers.HasLength(3),
                        "Expected 3 alarms - 1 on host and 2 deduced")
        self.assertThat(deduce_alarms_1, matchers.HasLength(1),
                        "Deduced alarm not found")
        self.assertThat(deduce_alarms_2, matchers.HasLength(1),
                        "Deduced alarm not found")

    def _validate_set_state(self, topology, instances):
        self.assertThat(topology, IsNotEmpty(), 'The topology graph is empty')
        host = g_utils.all_matches(
            topology,
            vitrage_type=NOVA_HOST_DATASOURCE,
            id=self._get_hostname(),
            vitrage_state=OperationalResourceState.ERROR,
            vitrage_aggregated_state=OperationalResourceState.ERROR)

        vm1 = g_utils.all_matches(
            topology,
            vitrage_type=NOVA_INSTANCE_DATASOURCE,
            id=instances[0].id,
            vitrage_state=OperationalResourceState.SUBOPTIMAL,
            vitrage_aggregated_state=OperationalResourceState.SUBOPTIMAL)

        vm2 = g_utils.all_matches(
            topology,
            vitrage_type=NOVA_INSTANCE_DATASOURCE,
            id=instances[1].id,
            vitrage_state=OperationalResourceState.SUBOPTIMAL,
            vitrage_aggregated_state=OperationalResourceState.SUBOPTIMAL)

        self.assertThat(host, matchers.HasLength(1))
        self.assertThat(vm1, matchers.HasLength(1))
        self.assertThat(vm2, matchers.HasLength(1))

    @staticmethod
    def _get_hostname():
        host = vitrage_utils.get_first_host()
        return host.get(VProps.ID)

    @staticmethod
    def _clean_timestamps(alist):
        try:
            del alist[5][1][0][VProps.VITRAGE_SAMPLE_TIMESTAMP]
            del alist[5][1][0][VProps.UPDATE_TIMESTAMP]
        except Exception:
            pass
        return alist



# Copyright 2017 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import time

from oslo_log import log as logging

from vitrage_tempest_plugin.tests.base import IsEmpty
from vitrage_tempest_plugin.tests.common.constants import DOCTOR_DATASOURCE
from vitrage_tempest_plugin.tests.common.constants import EntityCategory
from vitrage_tempest_plugin.tests.common.constants import VertexProperties \
    as VProps
from vitrage_tempest_plugin.tests.common.constants import VITRAGE_DATASOURCE
from vitrage_tempest_plugin.tests.common import general_utils as g_utils
from vitrage_tempest_plugin.tests.common.tempest_clients import TempestClients
from vitrage_tempest_plugin.tests.common import vitrage_utils as v_utils
from vitrage_tempest_plugin.tests.e2e.test_actions_base import TestActionsBase
from vitrage_tempest_plugin.tests import utils

LOG = logging.getLogger(__name__)

TRIGGER_ALARM_1 = 'e2e.test_overlapping_actions.trigger.alarm1'
TRIGGER_ALARM_2 = 'e2e.test_overlapping_actions.trigger.alarm2'
TRIGGER_ALARM_3 = 'e2e.test_overlapping_actions.trigger.alarm3'
TRIGGER_ALARM_4 = 'e2e.test_overlapping_actions.trigger.alarm4'
DEDUCED = 'e2e.test_overlapping_actions.deduced.alarm'

TRIGGER_ALARM_1_PROPS = {
    VProps.NAME: TRIGGER_ALARM_1,
    VProps.VITRAGE_CATEGORY: EntityCategory.ALARM,
    VProps.VITRAGE_TYPE: DOCTOR_DATASOURCE,
}

TRIGGER_ALARM_2_PROPS = {
    VProps.NAME: TRIGGER_ALARM_2,
    VProps.VITRAGE_CATEGORY: EntityCategory.ALARM,
    VProps.VITRAGE_TYPE: DOCTOR_DATASOURCE,
}

DEDUCED_PROPS = {
    VProps.NAME: DEDUCED,
    VProps.VITRAGE_CATEGORY: EntityCategory.ALARM,
    VProps.VITRAGE_TYPE: VITRAGE_DATASOURCE,
}


class TestOverlappingActions(TestActionsBase):

    @classmethod
    def setUpClass(cls):
        super(TestOverlappingActions, cls).setUpClass()
        cls._template = v_utils.add_template(
            'e2e_test_overlapping_actions.yaml')

    @classmethod
    def tearDownClass(cls):
        if cls._template is not None:
            v_utils.delete_template(cls._template['uuid'])

    @utils.tempest_logger
    def test_overlapping_action_set_state(self):
        try:
            # Do - first
            self._trigger_do_action(TRIGGER_ALARM_1)
            curr_host = v_utils.get_first_host()
            self.assertEqual(
                'ERROR',
                curr_host.get(VProps.VITRAGE_AGGREGATED_STATE),
                'state should change after set_state action')

            # Do - second
            self._trigger_do_action(TRIGGER_ALARM_2)
            curr_host = v_utils.get_first_host()
            self.assertEqual(
                'ERROR',
                curr_host.get(VProps.VITRAGE_AGGREGATED_STATE),
                'state should remain unchanged')

            # Undo - first
            self._trigger_undo_action(TRIGGER_ALARM_1)
            curr_host = v_utils.get_first_host()
            self.assertEqual(
                'ERROR',
                curr_host.get(VProps.VITRAGE_AGGREGATED_STATE),
                'state should remain unchanged')

            # Undo - second
            self._trigger_undo_action(TRIGGER_ALARM_2)
            curr_host = v_utils.get_first_host()
            self.assertEqual(
                self.orig_host.get(VProps.VITRAGE_AGGREGATED_STATE),
                curr_host.get(VProps.VITRAGE_AGGREGATED_STATE),
                'state should change after undo set_state action')
        finally:
            self._trigger_undo_action(TRIGGER_ALARM_1)
            self._trigger_undo_action(TRIGGER_ALARM_2)

    @utils.tempest_logger
    def test_overlapping_action_mark_down(self):
        try:
            host_name = self.orig_host.get(VProps.NAME)

            # Do - first
            self._trigger_do_action(TRIGGER_ALARM_3)
            nova_service = TempestClients.nova().services.list(
                host=host_name, binary='nova-compute')[0]
            self.assertEqual("down", nova_service.state)

            # Do - second
            self._trigger_do_action(TRIGGER_ALARM_4)
            nova_service = TempestClients.nova().services.list(
                host=host_name, binary='nova-compute')[0]
            self.assertEqual("down", nova_service.state)

            # Undo - first
            self._trigger_undo_action(TRIGGER_ALARM_3)
            nova_service = TempestClients.nova().services.list(
                host=host_name, binary='nova-compute')[0]
            self.assertEqual("down", nova_service.state)

            # Undo - second
            self._trigger_undo_action(TRIGGER_ALARM_4)
            nova_service = TempestClients.nova().services.list(
                host=host_name, binary='nova-compute')[0]
            self.assertEqual("up", nova_service.state)
        finally:
            self._trigger_undo_action(TRIGGER_ALARM_3)
            self._trigger_undo_action(TRIGGER_ALARM_4)
            # nova.host datasource may take up to snapshot_interval to update
            time.sleep(130)

    @utils.tempest_logger
    def test_overlapping_action_deduce_alarm(self):
        try:
            host_id = self.orig_host.get(VProps.VITRAGE_ID)

            # Do - first
            self._trigger_do_action(TRIGGER_ALARM_1)
            self._check_deduced(1, DEDUCED_PROPS, host_id)

            # Do - second
            self._trigger_do_action(TRIGGER_ALARM_2)
            self._check_deduced(1, DEDUCED_PROPS, host_id)

            # Undo - first
            self._trigger_undo_action(TRIGGER_ALARM_1)
            self._check_deduced(1, DEDUCED_PROPS, host_id)

            # Undo - second
            self._trigger_undo_action(TRIGGER_ALARM_2)
            self._check_deduced(0, DEDUCED_PROPS, host_id)
        finally:
            self._trigger_undo_action(TRIGGER_ALARM_1)
            self._trigger_undo_action(TRIGGER_ALARM_2)

    @utils.tempest_logger
    def test_overlapping_action_add_causal_relationship(self):
        try:
            # ---- Do first & second ----
            self._trigger_do_action(TRIGGER_ALARM_1)
            self._trigger_do_action(TRIGGER_ALARM_2)
            alarms = self.vitrage_client.alarm.list(
                vitrage_id=self.orig_host.get(VProps.VITRAGE_ID),
                all_tenants=True)

            deduced = g_utils.first_match(alarms, **DEDUCED_PROPS)
            trigger1 = g_utils.first_match(alarms, **TRIGGER_ALARM_1_PROPS)
            trigger2 = g_utils.first_match(alarms, **TRIGGER_ALARM_2_PROPS)

            # Get Rca for the deduced
            rca = self.vitrage_client.rca.get(deduced[VProps.VITRAGE_ID],
                                              all_tenants=True)
            self._check_rca(rca, [deduced, trigger1, trigger2], DEDUCED_PROPS)

            # Get Rca for trigger 1
            rca = self.vitrage_client.rca.get(trigger1[VProps.VITRAGE_ID],
                                              all_tenants=True)
            self._check_rca(rca, [deduced, trigger1], TRIGGER_ALARM_1_PROPS)

            # Get Rca for trigger 2
            rca = self.vitrage_client.rca.get(trigger2[VProps.VITRAGE_ID],
                                              all_tenants=True)
            self._check_rca(rca, [deduced, trigger2], TRIGGER_ALARM_2_PROPS)

            # ---- Undo - first ----
            self._trigger_undo_action(TRIGGER_ALARM_1)
            alarms = self.vitrage_client.alarm.list(
                vitrage_id=self.orig_host.get(VProps.VITRAGE_ID),
                all_tenants=True)

            deduced = g_utils.first_match(alarms, **DEDUCED_PROPS)
            trigger2 = g_utils.first_match(alarms, **TRIGGER_ALARM_2_PROPS)

            # Get Rca for the deduced
            rca = self.vitrage_client.rca.get(deduced[VProps.VITRAGE_ID],
                                              all_tenants=True)
            self._check_rca(rca, [deduced, trigger2], DEDUCED_PROPS)

            # Get Rca for trigger 2
            rca = self.vitrage_client.rca.get(trigger2[VProps.VITRAGE_ID],
                                              all_tenants=True)
            self._check_rca(rca, [deduced, trigger2], TRIGGER_ALARM_2_PROPS)

            # ---- Undo - second ----
            self._trigger_undo_action(TRIGGER_ALARM_2)
            alarms = self.vitrage_client.alarm.list(
                vitrage_id=self.orig_host.get(VProps.VITRAGE_ID),
                all_tenants=True)
            self.assertThat(
                g_utils.all_matches(alarms, **TRIGGER_ALARM_1_PROPS),
                IsEmpty(),
                'trigger alarm 1 should have been removed')
            self.assertThat(
                g_utils.all_matches(alarms, **TRIGGER_ALARM_2_PROPS),
                IsEmpty(),
                'trigger alarm 2 should have been removed')
            self.assertThat(
                g_utils.all_matches(alarms, **DEDUCED_PROPS),
                IsEmpty(),
                'deduced alarm should have been removed')
        finally:
            self._trigger_undo_action(TRIGGER_ALARM_1)
            self._trigger_undo_action(TRIGGER_ALARM_2)



# Copyright 2016 Nokia
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import json

from oslo_log import log as logging
from testtools import ExpectedException

from vitrage_tempest_plugin.tests.api.topology.base import BaseTopologyTest
from vitrage_tempest_plugin.tests.base import IsEmpty
import vitrage_tempest_plugin.tests.utils as utils
from vitrageclient.exceptions import ClientException

LOG = logging.getLogger(__name__)
NOVA_QUERY = '{"and": [{"==": {"vitrage_category": "RESOURCE"}},' \
             '{"==": {"vitrage_is_deleted": false}},' \
             '{"==": {"vitrage_is_placeholder": false}},' \
             '{"or": [{"==": {"vitrage_type": "openstack.cluster"}},' \
             '{"==": {"vitrage_type": "nova.instance"}},' \
             '{"==": {"vitrage_type": "nova.host"}},' \
             '{"==": {"vitrage_type": "nova.zone"}}]}]}'


class TestTopology(BaseTopologyTest):
    """Topology test class for Vitrage API tests."""

    NUM_INSTANCE = 3
    NUM_VOLUME = 1

    def tearDown(self):
        super(TestTopology, self).tearDown()
        self._rollback_to_default()

    @utils.tempest_logger
    def test_compare_api_and_cli(self):
        """compare_api_and_cli

        This test validate correctness of topology graph:
         cli via api
        """
        api_graph = self.vitrage_client.topology.get(all_tenants=True)
        cli_graph = utils.run_vitrage_command(
            'vitrage topology show --all-tenants')

        self._assert_graph_equal(api_graph, json.loads(cli_graph))

    @utils.tempest_logger
    def test_default_graph_all_tenants(self):
        """default_graph

        This test validate correctness of default topology graph, using the
        --all-tenants option.
        """
        self._do_test_default_graph(
            num_default_networks=self.num_default_networks, all_tenants=True)

    @utils.tempest_logger
    def test_default_graph(self):
        """default_graph

        This test validate correctness of default topology graph, not using the
        --all-tenants option. Since the tenant of all entities in the graph is
        either admin or None, this test is expected to be identical to the one
        with --all-tenants
        """
        num_admin_networks = self._calc_num_admin_tenant_networks()

        self._do_test_default_graph(num_default_networks=num_admin_networks,
                                    all_tenants=False)

    @utils.tempest_logger
    def test_default_graph_for_tenant(self):
        """default_graph

        This test validate correctness of default topology graph when queried
        by another tenant.
        """
        # Action - create entities as the default tenant
        self._create_entities(num_instances=self.NUM_INSTANCE,
                              num_volumes=self.NUM_VOLUME)

        # Calculate expected results for another tenant - the other tenant
        # should not see instances and volumes of the default tenant.
        # All it can see is its private network.
        tenant_client = self.vitrage_client_for_demo_user
        api_graph = tenant_client.topology.get(all_tenants=False)
        graph = self._create_graph_from_graph_dictionary(api_graph)

        entities = self._entities_validation_data(
            cluster_entities=0, cluster_edges=0, zone_entities=0,
            zone_edges=0, host_entities=0, host_edges=0,
            instance_entities=0, instance_edges=0, volume_entities=0,
            volume_edges=0)
        num_entities = self._calc_num_tenant_networks()

        # Test Assertions
        self._validate_graph_correctness(graph, num_entities, 0, entities)

    @utils.tempest_logger
    def test_graph_with_query(self):
        """graph_with_query

        This test validate correctness of topology graph
        with query
        """
        # Action
        self._create_entities(num_instances=self.NUM_INSTANCE,
                              num_volumes=self.NUM_VOLUME)

        # Calculate expected results
        api_graph = self.vitrage_client.topology.get(
            query=self._graph_query(),
            all_tenants=True)
        graph = self._create_graph_from_graph_dictionary(api_graph)
        entities = self._entities_validation_data(
            host_entities=1,
            host_edges=self.NUM_INSTANCE + 1,
            instance_entities=self.NUM_INSTANCE,
            instance_edges=self.NUM_INSTANCE)
        num_entities = self.num_default_entities + self.NUM_INSTANCE
        num_edges = self.num_default_edges + self.NUM_INSTANCE

        # Test Assertions
        self._validate_graph_correctness(graph,
                                         num_entities,
                                         num_edges,
                                         entities)

    @utils.tempest_logger
    def test_graph_with_query_for_tenant(self):
        """graph_with_query

        This test validate correctness of topology graph
        with query, when queried by another tenant.
        """
        # Action - create entities as the default tenant
        self._create_entities(num_instances=self.NUM_INSTANCE,
                              num_volumes=self.NUM_VOLUME)

        # Calculate expected results - the other tenant should not see
        # instances and volumes of the default tenant
        tenant_client = self.vitrage_client_for_demo_user
        api_graph = tenant_client.topology.get(query=self._graph_query())
        graph = self._create_graph_from_graph_dictionary(api_graph)

        entities = self._entities_validation_data(
            cluster_entities=0, cluster_edges=0, zone_entities=0,
            zone_edges=0, host_entities=0, host_edges=0,
            instance_entities=0, instance_edges=0, volume_entities=0,
            volume_edges=0)

        # Test Assertions
        self._validate_graph_correctness(graph, 0, 0, entities)

    @utils.tempest_logger
    def test_nova_tree(self):
        """nova_tree

        This test validate correctness of topology tree
        """
        # Action
        self._create_entities(num_instances=self.NUM_INSTANCE,
                              num_volumes=self.NUM_VOLUME)

        # Calculate expected results
        api_graph = self.vitrage_client.topology.get(
            graph_type='tree', query=NOVA_QUERY, all_tenants=True)
        graph = self._create_graph_from_tree_dictionary(api_graph)
        entities = self._entities_validation_data(
            host_entities=1,
            host_edges=self.NUM_INSTANCE + 1,
            instance_entities=self.NUM_INSTANCE,
            instance_edges=self.NUM_INSTANCE)
        num_entities = self.num_default_entities + self.NUM_INSTANCE
        num_edges = self.num_default_edges + self.NUM_INSTANCE

        # Test Assertions
        self._validate_graph_correctness(graph,
                                         num_entities,
                                         num_edges,
                                         entities)

    @utils.tempest_logger
    def test_tree_with_query(self):
        """tree_with_query

        This test validate correctness of topology tree
        with query
        """
        # Action
        self._create_entities(num_instances=self.NUM_INSTANCE)

        # Calculate expected results
        api_graph = self.vitrage_client.topology.get(
            graph_type='tree', query=self._tree_query(), all_tenants=True)
        graph = self._create_graph_from_tree_dictionary(api_graph)
        entities = self._entities_validation_data(
            host_entities=1, host_edges=1)

        # Test Assertions
        self._validate_graph_correctness(graph,
                                         self.num_default_entities,
                                         self.num_default_edges,
                                         entities)

    @utils.tempest_logger
    def test_tree_with_depth_exclude_instance(self):
        """tree_with_query

        This test validate correctness of topology tree
        with query
        """
        # Action
        self._create_entities(num_instances=self.NUM_INSTANCE)

        # Calculate expected results
        api_graph = self.vitrage_client.topology.get(
            limit=2, graph_type='tree', query=NOVA_QUERY, all_tenants=True)
        graph = self._create_graph_from_tree_dictionary(api_graph)
        entities = self._entities_validation_data(
            host_entities=1, host_edges=1)

        # Test Assertions
        self._validate_graph_correctness(graph,
                                         self.num_default_entities,
                                         self.num_default_edges,
                                         entities)

    @utils.tempest_logger
    def test_tree_with_depth_include_instance(self):
        """tree_with_query

        This test validate correctness of topology tree
        with query
        """
        # Action
        self._create_entities(num_instances=self.NUM_INSTANCE)

        # Calculate expected results
        api_graph = self.vitrage_client.topology.get(
            limit=3, graph_type='tree', query=NOVA_QUERY, all_tenants=True)
        graph = self._create_graph_from_tree_dictionary(api_graph)
        entities = self._entities_validation_data(
            host_entities=1,
            host_edges=self.NUM_INSTANCE + 1,
            instance_entities=self.NUM_INSTANCE,
            instance_edges=self.NUM_INSTANCE)
        num_entities = self.num_default_entities + self.NUM_INSTANCE
        num_edges = self.num_default_edges + self.NUM_INSTANCE

        # Test Assertions
        self._validate_graph_correctness(graph,
                                         num_entities,
                                         num_edges,
                                         entities)

    @utils.tempest_logger
    def test_graph_with_depth_and_no_root(self):
        """graph_with_depth_and_no_root

        This test validate correctness of topology
        graph with depth and without root
        """
        # Action
        self._create_entities(num_instances=self.NUM_INSTANCE,
                              num_volumes=self.NUM_VOLUME)

        with ExpectedException(ClientException,
                               "Graph-type 'graph' "
                               "requires a 'root' with 'depth'"):

            self.vitrage_client.topology.get(
                limit=2,
                root=None,
                all_tenants=True)

    @utils.tempest_logger
    def test_graph_with_no_match_query(self):
        """graph_with_no_match_query

        This test validate correctness of topology graph
        with no match query
        """
        # Action
        self._create_entities(num_instances=self.NUM_INSTANCE,
                              num_volumes=self.NUM_VOLUME)

        # Calculate expected results
        api_graph = self.vitrage_client.topology.get(
            query=self._graph_no_match_query(), all_tenants=True)

        # Test Assertions
        self.assertThat(api_graph['nodes'],
                        IsEmpty(), 'num of vertex node')
        self.assertThat(api_graph['links'], IsEmpty(), 'num of edges')

    @utils.tempest_logger
    def test_tree_with_no_match_query(self):
        """tree_with_no_match_query

        This test validate correctness of topology tree
        with no match query
        """
        # Action
        self._create_entities(num_instances=self.NUM_INSTANCE)

        # Calculate expected results
        api_graph = self.vitrage_client.topology.get(
            graph_type='tree',
            query=self._tree_no_match_query(),
            all_tenants=True)

        # Test Assertions
        self.assert_is_empty(api_graph)

    @utils.tempest_logger
    def _do_test_default_graph(self, num_default_networks, all_tenants):
        """default_graph

        This test validate correctness of default topology graph
        """
        # Action
        self._create_entities(num_instances=self.NUM_INSTANCE,
                              num_volumes=self.NUM_VOLUME)

        # Calculate expected results
        api_graph = \
            self.vitrage_client.topology.get(all_tenants=all_tenants)
        graph = self._create_graph_from_graph_dictionary(api_graph)
        entities = self._entities_validation_data(
            host_entities=1,
            host_edges=self.NUM_INSTANCE + 1,
            instance_entities=self.NUM_INSTANCE,
            instance_edges=2 * self.NUM_INSTANCE + self.NUM_VOLUME,
            volume_entities=self.NUM_VOLUME,
            volume_edges=self.NUM_VOLUME)
        num_entities = self.num_default_entities + self.NUM_VOLUME + \
            2 * self.NUM_INSTANCE + num_default_networks + \
            self.num_default_ports
        num_edges = self.num_default_edges + 3 * self.NUM_INSTANCE + \
            self.NUM_VOLUME + self.num_default_ports

        # Test Assertions
        self._validate_graph_correctness(graph,
                                         num_entities,
                                         num_edges,
                                         entities)



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log as logging

from vitrage_tempest_plugin.tests import utils

from vitrage_tempest_plugin.tests.api.topology.base import BaseTopologyTest
from vitrage_tempest_plugin.tests.common.general_utils\
    import tempest_resources_dir
from vitrage_tempest_plugin.tests.common import glance_utils
from vitrage_tempest_plugin.tests.common import heat_utils

LOG = logging.getLogger(__name__)


class TestHeatStack(BaseTopologyTest):
    NUM_STACKS = 1

    def tearDown(self):
        super(TestHeatStack, self).tearDown()
        heat_utils.delete_all_stacks()

    @utils.tempest_logger
    def test_nested_heat_stack(self):
        self._test_heat_stack(nested=True,
                              tmpl_file='heat_nested_template.yaml')

    @utils.tempest_logger
    def test_heat_stack(self):
        self._test_heat_stack(nested=False, tmpl_file='heat_template.yaml')

    def _test_heat_stack(self, nested, tmpl_file):
        """heat stack test

        This test validate correctness topology graph with heat stack module
        """
        template_file = tempest_resources_dir() + '/heat/' + tmpl_file
        image = glance_utils.get_first_image()
        # Action
        heat_utils.create_stacks(self.NUM_STACKS, nested, template_file,
                                 image['name'])

        # Calculate expected results
        api_graph = self.vitrage_client.topology.get(all_tenants=True)
        graph = self._create_graph_from_graph_dictionary(api_graph)
        entities = self._entities_validation_data(
            host_entities=1,
            host_edges=1 + self.NUM_STACKS,
            instance_entities=self.NUM_STACKS,
            instance_edges=3 * self.NUM_STACKS,
            network_entities=self.num_default_networks,
            network_edges=self.num_default_ports + self.NUM_STACKS,
            port_entities=self.num_default_ports + self.NUM_STACKS,
            port_edges=self.num_default_ports + 2 * self.NUM_STACKS,
            stack_entities=self.NUM_STACKS,
            stack_edges=self.NUM_STACKS)
        num_entities = self.num_default_entities + 3 * self.NUM_STACKS + \
            self.num_default_networks + self.num_default_ports
        num_edges = self.num_default_edges + 4 * self.NUM_STACKS + \
            self.num_default_ports

        # Test Assertions
        self._validate_graph_correctness(graph,
                                         num_entities,
                                         num_edges,
                                         entities)



# Copyright 2017 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_log import log as logging
import time

from vitrage_tempest_plugin.tests.common.constants import EntityCategory
from vitrage_tempest_plugin.tests.common.constants import VertexProperties \
    as VProps
from vitrage_tempest_plugin.tests.common.constants import VITRAGE_DATASOURCE
from vitrage_tempest_plugin.tests.common import vitrage_utils as v_util
from vitrage_tempest_plugin.tests.e2e.test_actions_base import TestActionsBase
from vitrage_tempest_plugin.tests import utils

LOG = logging.getLogger(__name__)

TRIGGER_ALARM_1 = 'e2e.test_template_actions.trigger.alarm1'
DEDUCED = 'e2e.test_template_actions.deduced.alarm'

TRIGGER_ALARM_2 = 'e2e.test_template_actions.trigger.alarm2'
DEDUCED_2 = 'e2e.test_template_actions.deduced.alarm2'

TEST_TEMPLATE = 'e2e_test_template_actions.yaml'
TEST_TEMPLATE_2 = 'e2e_test_template_actions_2.yaml'

INFILE_NAME = 'e2e_test_template_actions'
INFILE_NAME_2 = 'e2e_test_template_actions_2'


FOLDER_PATH = 'templates/api/e2e_test_template'

DEDUCED_PROPS = {
    VProps.NAME: DEDUCED,
    VProps.VITRAGE_CATEGORY: EntityCategory.ALARM,
    VProps.VITRAGE_TYPE: VITRAGE_DATASOURCE,
}

DEDUCED_PROPS_2 = {
    VProps.NAME: DEDUCED_2,
    VProps.VITRAGE_CATEGORY: EntityCategory.ALARM,
    VProps.VITRAGE_TYPE: VITRAGE_DATASOURCE,
}


class TestTemplateActions(TestActionsBase):

    def __init__(self, *args, **kwds):
        super(TestTemplateActions, self).__init__(*args, **kwds)
        self.added_template = None

    def tearDown(self):
        super(TestTemplateActions, self).tearDown()
        time.sleep(10)
        self._trigger_undo_action(TRIGGER_ALARM_1)
        if self.added_template is not None:
            v_util.delete_template(self.added_template['uuid'])
            self.added_template = None

    @utils.tempest_logger
    def test_evaluator_reload_with_new_template(self):
        """Test reload new template e2e

        1. raise trigger alarm (template is not loaded yet)
        2. add the relevant template
        3. check action is executed
        This checks that the evaluators are reloaded and run on all existing
         vertices.
        """
        host_id = self.orig_host.get(VProps.VITRAGE_ID)
        self._trigger_do_action(TRIGGER_ALARM_1)
        self.added_template = v_util.add_template(TEST_TEMPLATE,
                                                  folder=FOLDER_PATH)
        time.sleep(2)
        self._check_deduced(1, DEDUCED_PROPS, host_id)

    @utils.tempest_logger
    def test_evaluator_reload_with_existing_template(self):
        """Test reload new template e2e

        1.add the relevant template
        2.raise trigger alarm.
        3. check action is executed
        This checks that new workers execute new template
        """
        host_id = self.orig_host.get(VProps.VITRAGE_ID)
        self.added_template = v_util.add_template(TEST_TEMPLATE,
                                                  folder=FOLDER_PATH)
        self._trigger_do_action(TRIGGER_ALARM_1)
        self._check_deduced(1, DEDUCED_PROPS, host_id)

    @utils.tempest_logger
    def test_evaluator_reload_with_new_template_v2(self):
        """Test reload new template e2e v2

        1. raise trigger alarm
        2. add the relevant template
        3. delete the template.
        4. check action - should be not active.
        This checks that the evaluators are reloaded and run on all existing
         vertices.
         Checks temporary worker that was added to delete template.
        """
        host_id = self.orig_host.get(VProps.VITRAGE_ID)

        self._trigger_do_action(TRIGGER_ALARM_1)
        self.added_template = v_util.add_template(TEST_TEMPLATE,
                                                  folder=FOLDER_PATH)
        time.sleep(2)
        self._check_deduced(1, DEDUCED_PROPS, host_id)
        v_util.delete_template(self.added_template['uuid'])
        self.added_template = None
        time.sleep(2)
        self._check_deduced(0, DEDUCED_PROPS, host_id)

    @utils.tempest_logger
    def test_evaluator_reload_with_existing_template_v2(self):
        """Test reload new template e2e v2

        1.add the relevant template
        2.delete the template
        2.raise trigger alarm.
        3. check no deduced alarm
        This checks that template deleted properly and no action executed.
        :return:
        """
        host_id = self.orig_host.get(VProps.VITRAGE_ID)
        self.added_template = v_util.add_template(TEST_TEMPLATE,
                                                  folder=FOLDER_PATH)
        v_util.delete_template(self.added_template['uuid'])
        self.added_template = None
        self._trigger_do_action(TRIGGER_ALARM_1)
        self._check_deduced(0, DEDUCED_PROPS, host_id)

    @utils.tempest_logger
    def test_evaluator_reload_with_multiple_new_template(self):
        """Test reload new template e2e

        1. raise trigger alarm (template is not loaded yet)
        2. add 2 new templates.
        3. check both actions are executed
        This checks that the evaluators are reloaded for both templates
         and run on all existing vertices.
        """
        second_template = None
        try:
            host_id = self.orig_host.get(VProps.VITRAGE_ID)
            self._trigger_do_action(TRIGGER_ALARM_1)
            self._trigger_do_action(TRIGGER_ALARM_2)
            v_util.add_template(folder=FOLDER_PATH)
            self.added_template = v_util.get_first_template(name=INFILE_NAME)
            second_template = v_util.get_first_template(name=INFILE_NAME_2)
            self._check_deduced(1, DEDUCED_PROPS, host_id)
            self._check_deduced(1, DEDUCED_PROPS_2, host_id)
        finally:
            if second_template:
                v_util.delete_template(second_template['uuid'])
            self._trigger_undo_action(TRIGGER_ALARM_1)
            self._trigger_undo_action(TRIGGER_ALARM_2)



# Copyright 2017 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import time

from oslo_log import log as logging
from oslo_serialization import jsonutils
import testtools
from testtools.matchers import HasLength

from vitrage_tempest_plugin.tests.api.event.base import BaseTestEvents
from vitrage_tempest_plugin.tests.common import nova_utils
from vitrage_tempest_plugin.tests.common.tempest_clients import TempestClients
from vitrage_tempest_plugin.tests.common import vitrage_utils as v_utils
from vitrage_tempest_plugin.tests import utils
from vitrage_tempest_plugin.tests.utils import wait_for_status

LOG = logging.getLogger(__name__)


WF_NAME = 'wf_for_tempest_test_1234'

WF_DEFINITION = """
---
version: '2.0'

wf_for_tempest_test_1234:
  type: direct
  input:
    - farewell

  tasks:
    goodbye:
      action: std.echo output="<% $.farewell %>, Tempest Test!"
"""


@testtools.skip("skip for now until mistral fixes the failures")
class TestMistralNotifier(BaseTestEvents):

    TRIGGER_ALARM_1 = "notifiers.mistral.trigger.alarm.1"
    TRIGGER_ALARM_2 = "notifiers.mistral.trigger.alarm.2"
    TRIGGER_ALARM_3 = "notifiers.mistral.trigger.alarm.3"
    TRIGGER_ALARM_FOR_FUNCTION = "notifiers.mistral.trigger.alarm.for.function"
    TRIGGER_ALARM_FOR_FUNCTION_v3 = \
        "notifiers.mistral.trigger.alarm.for.function.v3"

    @classmethod
    def setUpClass(cls):
        super(TestMistralNotifier, cls).setUpClass()
        cls.mistral_client = TempestClients.mistral()
        cls._templates = []
        cls._templates.append(v_utils.add_template('v1_execute_mistral.yaml'))
        cls._templates.append(v_utils.add_template('v2_execute_mistral.yaml'))
        cls._templates.append(v_utils.add_template('v3_execute_mistral.yaml'))

        # Create a Mistral workflow
        cls.mistral_client.workflows.create(WF_DEFINITION)

    @classmethod
    def tearDownClass(cls):
        if cls._templates is not None:
            v_utils.delete_template(cls._templates[0]['uuid'])
            v_utils.delete_template(cls._templates[1]['uuid'])
            v_utils.delete_template(cls._templates[2]['uuid'])

        # Delete the workflow
        cls.mistral_client.workflows.delete(WF_NAME)
        nova_utils.delete_all_instances()

    @utils.tempest_logger
    def test_execute_mistral_v1(self):
        self._do_test_execute_mistral(self.TRIGGER_ALARM_1)

    @utils.tempest_logger
    def test_execute_mistral_v2(self):
        self._do_test_execute_mistral(self.TRIGGER_ALARM_2)

    @utils.tempest_logger
    def test_execute_mistral_v3(self):
        self._do_test_execute_mistral(self.TRIGGER_ALARM_3)

    @utils.tempest_logger
    def test_execute_mistral_with_function_v2(self):
        # Execute the basic test
        self._do_test_execute_mistral(self.TRIGGER_ALARM_FOR_FUNCTION)
        self._do_test_function(self.TRIGGER_ALARM_FOR_FUNCTION)

    @utils.tempest_logger
    def test_execute_mistral_with_function_v3(self):
        # Execute the basic test
        self._do_test_execute_mistral(self.TRIGGER_ALARM_FOR_FUNCTION_v3)
        self._do_test_function(self.TRIGGER_ALARM_FOR_FUNCTION_v3)

    @utils.tempest_logger
    def test_execute_mistral_more_than_once(self):
        executions = self.mistral_client.executions.list()
        self.assertIsNotNone(executions,
                             'Failed to get the list of workflow executions')
        num_executions = len(executions)

        # Make sure there are at least two instances in the environment
        nova_utils.create_instances(num_instances=2, set_public_network=True)
        num_instances = len(TempestClients.nova().servers.list())

        # Add a template that executes the same Mistral workflow for every
        # instance. This should immediately trigger execute_mistral actions.
        template = None
        try:
            template = v_utils.add_template('v3_execute_mistral_twice.yaml')
        finally:
            if template:
                v_utils.delete_template(template['uuid'])   # no longer needed

        time.sleep(2)   # wait for the evaluator to process the new template

        # Verify that there is an execution for every instance
        executions = self.mistral_client.executions.list()
        self.assertIsNotNone(executions,
                             'Failed to get the list of workflow executions')

        msg = "There are %d executions. Expected number of executions: %d " \
              "(old number of executions) + %d (number of instances)" % \
              (len(executions), num_executions, num_instances)
        self.assertThat(executions, HasLength(num_executions + num_instances),
                        msg)

        executed_on_instances = set()
        for i in range(num_instances):
            # There may be many old executions in the list. The relevant ones
            # are at the end. Check the last `num_instances` executions.
            execution = \
                self.mistral_client.executions.get(executions[-i].id)
            execution_input = jsonutils.loads(execution.input)
            executed_on_instances.add(execution_input['farewell'])

        msg = "There are %d instances in the graph but only %d distinct " \
              "executions" % (num_instances, len(executed_on_instances))
        self.assertThat(executed_on_instances, HasLength(num_instances), msg)

    def _do_test_function(self, trigger):
        # Make sure that the workflow execution was done with the correct input
        # (can be checked even if the Vitrage alarm is already down)
        executions = self.mistral_client.executions.list()
        last_execution = executions[0]
        for execution in executions:
            if execution.updated_at > last_execution.updated_at:
                last_execution = execution
        execution_input_str = last_execution.input
        self.assertIsNotNone(execution_input_str,
                             'The last execution had no input')
        self.assertIn('farewell', execution_input_str,
                      'No \'farewell\' key in the last execution input')
        execution_input = jsonutils.loads(execution_input_str)
        farewell_value = execution_input['farewell']
        self.assertIsNotNone(farewell_value, '\'farewell\' input parameter is '
                                             'None in last workflow execution')
        self.assertEqual(trigger, farewell_value,
                         '\'farewell\' input parameter does not match the'
                         'alarm name')

    def _do_test_execute_mistral(self, trigger_alarm):
        executions = self.mistral_client.executions.list()
        self.assertIsNotNone(executions,
                             'Failed to get the list of workflow executions')
        num_executions = len(executions)

        alarms = utils.wait_for_answer(2, 0.5, self._check_alarms)
        self.assertIsNotNone(alarms, 'Failed to get the list of alarms')
        num_alarms = len(alarms)

        try:
            # Trigger an alarm. According to v1_execute_mistral.yaml template,
            # the alarm should cause execution of the workflow
            self._trigger_do_action(trigger_alarm)

            # Wait for the alarm to be raised
            self.assertTrue(wait_for_status(
                10,
                self._check_num_vitrage_alarms,
                num_alarms=num_alarms + 1),
                'Trigger alarm was not raised')

            # Wait for the Mistral workflow execution
            self.assertTrue(wait_for_status(
                20,
                self._check_mistral_workflow_execution,
                num_executions=num_executions + 1),
                'Mistral workflow was not executed')

        finally:
            self._rollback_to_default(trigger_alarm, num_alarms)

    def _rollback_to_default(self, trigger_alarm, num_alarms):
        # Clear the trigger alarm and wait it to be deleted
        self._trigger_undo_action(trigger_alarm)

        self.assertTrue(wait_for_status(
            10,
            self._check_num_vitrage_alarms,
            num_alarms=num_alarms),
            'Vitrage trigger alarm was not deleted')

    @classmethod
    def _check_num_vitrage_alarms(cls, num_alarms):
        vitrage_alarms = cls.vitrage_client.alarm.list(vitrage_id='all',
                                                       all_tenants=True)
        if len(vitrage_alarms) == num_alarms:
            return True
        return False

    def _check_mistral_workflow_execution(self, num_executions):
        if len(self.mistral_client.executions.list()) == num_executions:
            return True
        return False



# Copyright 2016 Nokia
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from datetime import datetime
from itertools import chain
from itertools import islice

import networkx as nx
from networkx.readwrite import json_graph

from oslo_log import log as logging
from oslo_utils import timeutils
from tempest.common import credentials_factory as common_creds
from tempest import test
from testtools.matchers import HasLength
from testtools.matchers import Not

from vitrage_tempest_plugin.tests.common.constants import AODH_DATASOURCE
from vitrage_tempest_plugin.tests.common.constants import \
    CINDER_VOLUME_DATASOURCE
from vitrage_tempest_plugin.tests.common.constants import EdgeProperties
from vitrage_tempest_plugin.tests.common.constants import EntityCategory
from vitrage_tempest_plugin.tests.common.constants import HEAT_STACK_DATASOURCE
from vitrage_tempest_plugin.tests.common.constants import \
    NEUTRON_NETWORK_DATASOURCE
from vitrage_tempest_plugin.tests.common.constants import \
    NEUTRON_PORT_DATASOURCE
from vitrage_tempest_plugin.tests.common.constants import NOVA_HOST_DATASOURCE
from vitrage_tempest_plugin.tests.common.constants import \
    NOVA_INSTANCE_DATASOURCE
from vitrage_tempest_plugin.tests.common.constants import NOVA_ZONE_DATASOURCE
from vitrage_tempest_plugin.tests.common.constants import OPENSTACK_CLUSTER
from vitrage_tempest_plugin.tests.common.constants import VertexProperties \
    as VProps
from vitrage_tempest_plugin.tests.common import general_utils
from vitrage_tempest_plugin.tests.common.tempest_clients import TempestClients
from vitrage_tempest_plugin.tests import utils

import warnings

LOG = logging.getLogger(__name__)

IsEmpty = lambda: HasLength(0)
IsNotEmpty = lambda: Not(IsEmpty())

_MAX_LENGTH = 80


def unorderable_list_difference(expected, actual, ignore_duplicate=False):
    missing = []
    unexpected = []
    while expected:
        item = expected.pop()
        try:
            actual.remove(item)
        except ValueError:
            missing.append(item)
        if ignore_duplicate:
            for lst in expected, actual:
                try:
                    while True:
                        lst.remove(item)
                except ValueError:
                    pass
    if ignore_duplicate:
        while actual:
            item = actual.pop()
            unexpected.append(item)
            try:
                while True:
                    actual.remove(item)
            except ValueError:
                pass
        return missing, unexpected

    # anything left in actual is unexpected
    return missing, actual


def safe_repr(obj, short=False):
    try:
        result = repr(obj)
    except Exception:
        result = object.__repr__(obj)
    if not short or len(result) < _MAX_LENGTH:
        return result
    return result[:_MAX_LENGTH] + ' [truncated]...'


class BaseVitrageTempest(test.BaseTestCase):
    """Base test class for All Vitrage tests."""

    credentials = ['primary']

    NUM_VERTICES_PER_TYPE = 'num_vertices'
    NUM_EDGES_PER_TYPE = 'num_edges_per_type'
    # TODO(e0ne): use credentials from the config
    DEMO_PROJECT_NAME = 'demo'

    def assert_list_equal(self, l1, l2, message=None):
        self.assertListEqual(l1, l2, message)

    def assert_dict_equal(self, d1, d2, message=None):
        self.assertDictEqual(d1, d2, message)

    def assert_set_equal(self, s1, s2, message=None):
        self.assertSetEqual(s1, s2, message)

    def assert_sequence_equal(self, s1, s2, message=None):
        self.assertSequenceEqual(s1, s2, message)

    def assert_tuple_equal(self, t1, t2, message=None):
        self.assertTupleEqual(t1, t2, message)

    def assert_items_equal(self, s1, s2, message=None):
        self.assertItemsEqual(s1, s2, message)

    def assertItemsEqual(self, expected_seq, actual_seq, msg=None):
        try:
            expected = sorted(expected_seq)
            actual = sorted(actual_seq)
        except TypeError:
            # Unsortable items (example: set(), complex(), ...)
            expected = list(expected_seq)
            actual = list(actual_seq)
            missing, unexpected = unorderable_list_difference(
                expected, actual, ignore_duplicate=False
            )
        else:
            return self.assertSequenceEqual(expected, actual, msg=msg)

        errors = []
        if missing:
            errors.append('Expected, but missing:\n    %s' %
                          safe_repr(missing))
        if unexpected:
            errors.append('Unexpected, but present:\n    %s' %
                          safe_repr(unexpected))
        if errors:
            standard_msg = '\n'.join(errors)
            self.fail(self._formatMessage(msg, standard_msg))

    def assert_timestamp_equal(self, first, second, msg=None):
        """Checks that two timestamps are equals.

        This relies on assertAlmostEqual to avoid rounding problem, and only
        checks up the first microsecond values.

        """
        return self.assertAlmostEqual(timeutils.delta_seconds(first, second),
                                      0.0,
                                      places=5, msg=msg)

    def assert_is_empty(self, obj):
        try:
            if len(obj) != 0:
                self.fail("%s is not empty" % type(obj))
        except (TypeError, AttributeError):
            self.fail("%s doesn't have length" % type(obj))

    def assert_is_not_empty(self, obj):
        try:
            if len(obj) == 0:
                self.fail("%s is empty" % type(obj))
        except (TypeError, AttributeError):
            self.fail("%s doesn't have length" % type(obj))

    def setUp(self):
        super(BaseVitrageTempest, self).setUp()
        warnings.filterwarnings(action="ignore",
                                message="unclosed",
                                category=ResourceWarning)
        logger = logging.getLogger('urllib3.connectionpool').logger
        logger.setLevel(logging.INFO)

    def tearDown(self):
        super(BaseVitrageTempest, self).tearDown()
        warnings.filterwarnings(action="ignore",
                                message="unclosed",
                                category=ResourceWarning)

    # noinspection PyPep8Naming
    @classmethod
    def setUpClass(cls):
        super(BaseVitrageTempest, cls).setUpClass()
        warnings.filterwarnings(action="ignore",
                                message="unclosed",
                                category=ResourceWarning)
        TempestClients.class_init(cls.os_primary.credentials)
        cls.vitrage_client = TempestClients.vitrage()
        cls.vitrage_client_for_demo_user = \
            TempestClients.vitrage_client_for_user()

        cls.num_default_networks = \
            len(TempestClients.neutron().list_networks()['networks'])
        cls.num_default_ports = cls._get_num_default_ports()
        cls.num_default_entities = 3
        cls.num_default_edges = 2
        cls.num_demo_tenant_networks = cls._calc_num_tenant_networks()

    @classmethod
    def setup_credentials(cls):
        # Do not create new tempest network resources
        # We use the default networks that come in devstack
        # see the set_network_resources doc in tempest base
        cls.set_network_resources()
        super(BaseVitrageTempest, cls).setup_credentials()

    @classmethod
    def _get_num_default_ports(cls):
        ports = TempestClients.neutron().list_ports()['ports']
        nova_ports = general_utils.all_matches(ports,
                                               device_owner='compute:nova')
        LOG.debug('ports: %s, nova_ports: %s', ports, nova_ports)
        return len(nova_ports)

    def _assert_graph_equal(self, g1, g2, msg=''):
        """Checks that two graphs are equals.

        This relies on assert_dict_equal when comparing the nodes and the
        edges of each graph.
        """
        g1_nodes = g1['nodes']
        g1_links = g1['links']

        g2_nodes = g2['nodes']
        g2_links = g1['links']

        to_remove = {'vitrage_sample_timestamp',
                     'update_timestamp',
                     'graph_index'}

        self._delete_keys_from_dicts(chain(g1_nodes, g2_nodes), to_remove)

        self.assert_items_equal(g1_nodes, g2_nodes,
                                '%s Nodes of each graph are not equal' % msg)
        self.assert_items_equal(g1_links, g2_links,
                                '%s Edges of each graph are not equal' % msg)

    @staticmethod
    def _delete_keys_from_dicts(dictionaries, keys_to_remove):
        for dictionary in dictionaries:
            for key in keys_to_remove:
                if key in dictionary:
                    del dictionary[key]

    def _entities_validation_data(self, **kwargs):
        validation_data = []

        # openstack.cluster
        props = {VProps.VITRAGE_CATEGORY: EntityCategory.RESOURCE,
                 VProps.VITRAGE_TYPE: OPENSTACK_CLUSTER,
                 self.NUM_VERTICES_PER_TYPE: kwargs.get('cluster_entities', 1),
                 self.NUM_EDGES_PER_TYPE: kwargs.get('cluster_edges', 1)}
        validation_data.append(props)

        # nova.zone
        props = {VProps.VITRAGE_CATEGORY: EntityCategory.RESOURCE,
                 VProps.VITRAGE_TYPE: NOVA_ZONE_DATASOURCE,
                 self.NUM_VERTICES_PER_TYPE: kwargs.get('zone_entities', 1),
                 self.NUM_EDGES_PER_TYPE: kwargs.get('zone_edges', 2)}
        validation_data.append(props)

        # nova.host
        props = {VProps.VITRAGE_CATEGORY: EntityCategory.RESOURCE,
                 VProps.VITRAGE_TYPE: NOVA_HOST_DATASOURCE,
                 self.NUM_VERTICES_PER_TYPE: kwargs.get('host_entities', 1),
                 self.NUM_EDGES_PER_TYPE: kwargs.get('host_edges', 1)}
        validation_data.append(props)

        # nova.instance
        props = {VProps.VITRAGE_CATEGORY: EntityCategory.RESOURCE,
                 VProps.VITRAGE_TYPE: NOVA_INSTANCE_DATASOURCE,
                 self.NUM_VERTICES_PER_TYPE: kwargs.get(
                     'instance_entities', 0),
                 self.NUM_EDGES_PER_TYPE: kwargs.get(
                     'instance_edges', 0)}
        validation_data.append(props)

        # cinder.volume
        props = {VProps.VITRAGE_CATEGORY: EntityCategory.RESOURCE,
                 VProps.VITRAGE_TYPE: CINDER_VOLUME_DATASOURCE,
                 self.NUM_VERTICES_PER_TYPE: kwargs.get(
                     'volume_entities', 0),
                 self.NUM_EDGES_PER_TYPE: kwargs.get(
                     'volume_edges', 0)}
        validation_data.append(props)

        # switch
        props = {VProps.VITRAGE_CATEGORY: EntityCategory.RESOURCE,
                 VProps.VITRAGE_TYPE: 'switch',
                 self.NUM_VERTICES_PER_TYPE: kwargs.get(
                     'switch_entities', 0),
                 self.NUM_EDGES_PER_TYPE: kwargs.get(
                     'switch_edges', 0)}
        validation_data.append(props)

        # aodh
        props = {VProps.VITRAGE_CATEGORY: EntityCategory.ALARM,
                 VProps.VITRAGE_TYPE: AODH_DATASOURCE,
                 self.NUM_VERTICES_PER_TYPE: kwargs.get(
                     'aodh_entities', 0),
                 self.NUM_EDGES_PER_TYPE: kwargs.get(
                     'aodh_edges', 0)}
        validation_data.append(props)

        # neutron.network
        if kwargs.get('network_entities') is not None:
            props = {VProps.VITRAGE_CATEGORY: EntityCategory.RESOURCE,
                     VProps.VITRAGE_TYPE: NEUTRON_NETWORK_DATASOURCE,
                     self.NUM_VERTICES_PER_TYPE: kwargs.get(
                         'network_entities', 0),
                     self.NUM_EDGES_PER_TYPE: kwargs.get(
                         'network_edges', 0)}
            validation_data.append(props)

        # neutron.port
        if kwargs.get('port_entities') is not None:
            props = {VProps.VITRAGE_CATEGORY: EntityCategory.RESOURCE,
                     VProps.VITRAGE_TYPE: NEUTRON_PORT_DATASOURCE,
                     self.NUM_VERTICES_PER_TYPE: kwargs.get(
                         'port_entities', 0),
                     self.NUM_EDGES_PER_TYPE: kwargs.get(
                         'port_edges', 0)}
            validation_data.append(props)

        # heat.stack
        props = {VProps.VITRAGE_CATEGORY: EntityCategory.RESOURCE,
                 VProps.VITRAGE_TYPE: HEAT_STACK_DATASOURCE,
                 self.NUM_VERTICES_PER_TYPE: kwargs.get(
                     'stack_entities', 0),
                 self.NUM_EDGES_PER_TYPE: kwargs.get(
                     'stack_edges', 0)}
        validation_data.append(props)

        return validation_data

    def _create_graph_from_tree_dictionary(self,
                                           api_graph,
                                           graph=None,
                                           ancestor=None):
        children = []
        graph = nx.MultiDiGraph() if not graph else graph

        if 'children' in api_graph:
            children = api_graph['children']
            del api_graph['children']

        vitrage_id = api_graph[VProps.VITRAGE_ID]
        graph.add_node(vitrage_id, **api_graph)
        if ancestor:
            graph.add_edge(ancestor[VProps.VITRAGE_ID], vitrage_id, 'label')

        for entity in children:
            self._create_graph_from_tree_dictionary(entity, graph, api_graph)

        return graph

    def _create_graph_from_graph_dictionary(self, api_graph):
        self.assertIsNotNone(api_graph)
        graph = nx.MultiDiGraph()

        nodes = api_graph['nodes']
        for i, node in enumerate(nodes):
            graph.add_node(str(i), **node)

        edges = api_graph['links']
        for edge in edges:
            graph.add_edge(str(edge['source']),
                           str(edge['target']),
                           EdgeProperties.RELATIONSHIP_TYPE)

        return graph

    @staticmethod
    def _get_vertices(graph, _filter):

        def check_vertex(node):
            _, node_data = node
            for key, content in _filter.items():
                if not node_data.get(key) == content:
                    return False
            return True

        return list(filter(check_vertex, graph.nodes(data=True)))

    def _validate_graph_correctness(self,
                                    graph,
                                    num_entities,
                                    num_edges,
                                    entities):
        self.assertIsNotNone(graph)
        self.assertIsNotNone(entities)

        for entity in entities:
            vitrage_type = entity[VProps.VITRAGE_TYPE]
            vitrage_category = entity[VProps.VITRAGE_CATEGORY]
            query = {
                VProps.VITRAGE_CATEGORY: vitrage_category,
                VProps.VITRAGE_TYPE: vitrage_type,
                VProps.VITRAGE_IS_DELETED: False,
                VProps.VITRAGE_IS_PLACEHOLDER: False
            }
            entity_vertices = self._get_vertices(graph, _filter=query)

            expected_num_vertices = entity[self.NUM_VERTICES_PER_TYPE]
            observed_num_vertices = len(entity_vertices)
            failed_msg = ('Num entity_vertices is incorrect for: %s\n %s' %
                          (vitrage_type, self._to_dict(graph)))

            self.assertEqual(expected_num_vertices,
                             observed_num_vertices,
                             failed_msg)

            def num_of_edges_for(v_id):
                return len(graph.out_edges(v_id)) + len(graph.in_edges(v_id))

            # TODO(iafek): bug - edges between entities of the same type are
            # counted twice
            observed_entity_num_edges = sum(
                (num_of_edges_for(v_id) for v_id, _ in entity_vertices)
            )

            expected_entity_num_edges = entity[self.NUM_EDGES_PER_TYPE]
            failed_msg = ('Num edges is incorrect for: %s\n %s' %
                          (vitrage_type, self._to_dict(graph)))

            self.assertEqual(expected_entity_num_edges,
                             observed_entity_num_edges,
                             failed_msg)

        nodes = graph.nodes(data=True)
        vertices = self._extract_graph_vertices_data(nodes)
        self.assertEqual(num_entities, len(vertices), self._to_dict(graph))
        self.assertEqual(num_edges, len(graph.edges()), self._to_dict(graph))

        self._validate_timestamps(vertices)

    @staticmethod
    def _to_dict(graph):
        return json_graph.node_link_data(graph)

    # This will unzip the nodes and create a tuple of nodes with data only.
    # Using next and islice because zip returns iterator on py3
    # e.g. (id1, data1), (id2, data2) --> (data1, data2)
    @staticmethod
    def _extract_graph_vertices_data(nodes):
        def unzip(_nodes):
            return zip(*_nodes)

        return next(islice(unzip(nodes), 1, 2), ())

    def _validate_timestamps(self, vertices):
        self._validate_timestamp(vertices, VProps.UPDATE_TIMESTAMP)
        self._validate_timestamp(vertices, VProps.VITRAGE_SAMPLE_TIMESTAMP)

    def _validate_timestamp(self, vertices, timestamp_name):
        for vertex in vertices:
            timestamp = vertex.get(timestamp_name)
            if timestamp:
                try:
                    datetime.strptime(timestamp, utils.TIMESTAMP_FORMAT)
                except ValueError:
                    self.fail('Unexpected timestamp format of \'%s\' in: %s\n'
                              'The format should be: %s' %
                              (timestamp_name, vertex, utils.TIMESTAMP_FORMAT))

    @staticmethod
    def _calc_num_admin_tenant_networks():
        neutron_client = TempestClients.neutron()
        admin_creds = common_creds.get_configured_admin_credentials()
        tenant_networks = neutron_client.list_networks(
            tenant_id=admin_creds.project_id)['networks']
        return len(tenant_networks)

    @classmethod
    def _calc_num_tenant_networks(cls):
        neutron_client = TempestClients.neutron_client_for_user()
        tenant_networks = neutron_client.list_networks(
            tenant_id=cls.os_primary.credentials.project_id)['networks']
        return len(tenant_networks)

    @classmethod
    def _get_demo_tenant_id(cls):
        projects = TempestClients.keystone().projects.list()
        for project in projects:
            if cls.DEMO_PROJECT_NAME == project.name:
                return project.id
        return None



# Copyright 2017 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from os import path


def first_match(list_of_dicts, **kwargs):
    subset_dict = _remove_none_values(**kwargs)
    for d in list_of_dicts:
        if is_subset(subset_dict, d):
            return d


def all_matches(list_of_dicts, **kwargs):
    subset_dict = _remove_none_values(**kwargs)
    return [d for d in list_of_dicts if is_subset(subset_dict, d)]


def is_subset(subset, full):
    if not subset:
        return True
    full_dict = full
    if type(full) is not dict:
        full_dict = vars(full)
    return subset.items() <= full_dict.items()


def _remove_none_values(**kwargs):
    return {k: v for k, v in kwargs.items() if v is not None}


def tempest_resources_dir():
    return path.join(path.dirname(path.dirname(__file__)), 'resources')



# Copyright 2019 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log as logging

from vitrage_tempest_plugin.tests.api.templates.test_template_v2 \
    import TestTemplatesApis

LOG = logging.getLogger(__name__)

WITH_PARAMS_TEMPLATE = 'v3_with_params.yaml'
WITH_DEFAULT_PARAMS_TEMPLATE = 'v3_with_default_params.yaml'

TEMPLATE_STRING = """
metadata:
 version: 3
 name: template1
 description: simple template
 type: standard
entities:
 alarm:
  name: cpu problem
 host:
  type: nova.host
scenarios:
 - condition: alarm [ on ] host
   actions:
     - set_state:
        state: ERROR
        target: host
"""


class TestTemplatesV3(TestTemplatesApis):

    def test_template_validate_with_missing_parameters(self):
        self._validate_with_missing_parameters(WITH_PARAMS_TEMPLATE)

    def test_template_validate_with_missing_parameter(self):
        self._validate_with_missing_parameter(WITH_PARAMS_TEMPLATE)

    def test_template_validate_with_parameters(self):
        self._validate_with_parameters(WITH_PARAMS_TEMPLATE)

    def test_template_validate_with_default_parameters(self):
        self._validate_with_default_parameters(WITH_DEFAULT_PARAMS_TEMPLATE)

    def test_template_add_with_missing_parameters(self):
        self._add_with_missing_parameters(WITH_PARAMS_TEMPLATE)

    def test_template_add_with_missing_parameter(self):
        self._add_with_missing_parameter(WITH_PARAMS_TEMPLATE)

    def test_template_add_with_parameters(self):
        self._add_with_parameters(WITH_PARAMS_TEMPLATE)

    def test_template_add_with_default_parameters(self):
        self._add_with_default_parameters(WITH_DEFAULT_PARAMS_TEMPLATE)

    def test_template_add_by_string(self):
        self._add_by_string(TEMPLATE_STRING)



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log as logging

from vitrage_tempest_plugin.tests.api.rca.base import BaseRcaTest
from vitrage_tempest_plugin.tests.api.rca.base import RCA_ALARM_NAME
from vitrage_tempest_plugin.tests.common import nova_utils
from vitrage_tempest_plugin.tests.common import vitrage_utils as v_utils
from vitrage_tempest_plugin.tests import utils

LOG = logging.getLogger(__name__)


class TestRca(BaseRcaTest):
    """RCA test class for Vitrage API tests."""

    def tearDown(self):
        super(TestRca, self).tearDown()
        self._clean_all()

    @classmethod
    def setUpClass(cls):
        super(TestRca, cls).setUpClass()
        cls._template = v_utils.add_template('host_aodh_alarm_for_rca.yaml')

    # noinspection PyPep8Naming
    @classmethod
    def tearDownClass(cls):
        super(TestRca, cls).tearDownClass()
        if cls._template is not None:
            v_utils.delete_template(cls._template['uuid'])

    @utils.tempest_logger
    def test_validate_deduce_alarms(self):
        """validate_deduce_alarms test

        There tests validates correctness of deduce alarms
        (created by special template file), and equals there
        resource_id with created instances id
        """
        instances = nova_utils.create_instances(num_instances=2,
                                                set_public_network=True)
        self._create_alarm(
            resource_id=self._get_hostname(),
            alarm_name=RCA_ALARM_NAME)
        api_alarms = self.vitrage_client.alarm.list(vitrage_id='all',
                                                    all_tenants=True)

        self._validate_deduce_alarms(alarms=api_alarms,
                                     instances=instances)

    @utils.tempest_logger
    def test_validate_set_state(self):
        """validate_set_state test

        There tests validates correctness of topology resource
        state, after alarms creation (by special template file),
        source state - ERROR
        target state - SUBOPTIMAL (caused 2 created instance)
        """
        instances = nova_utils.create_instances(num_instances=2,
                                                set_public_network=True)
        self._create_alarm(
            resource_id=self._get_hostname(),
            alarm_name=RCA_ALARM_NAME)
        topology = self.vitrage_client.topology.get(all_tenants=True)

        self._validate_set_state(topology=topology['nodes'],
                                 instances=instances)



# Copyright 2017 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import time

from oslo_log import log as logging
from testtools import matchers

from vitrage_tempest_plugin.tests.base import BaseVitrageTempest
from vitrage_tempest_plugin.tests.common.constants import VertexProperties as \
    VProps
from vitrage_tempest_plugin.tests.common import general_utils as g_utils
from vitrage_tempest_plugin.tests.common import vitrage_utils

LOG = logging.getLogger(__name__)


class TestActionsBase(BaseVitrageTempest):
    @classmethod
    def setUpClass(cls):
        super(TestActionsBase, cls).setUpClass()
        host = vitrage_utils.get_first_host()
        if not host:
            raise Exception("No host found")
        if not host.get(VProps.VITRAGE_AGGREGATED_STATE) == 'AVAILABLE':
            raise Exception("Host is not running %s" % host)
        cls.orig_host = host

    def _trigger_do_action(self, trigger_name):
        vitrage_utils.generate_fake_host_alarm(
            self.orig_host.get('name'),
            enabled=True,
            event_type=trigger_name
        )
        time.sleep(2)

    def _trigger_undo_action(self, trigger_name):
        vitrage_utils.generate_fake_host_alarm(
            self.orig_host.get('name'),
            enabled=False,
            event_type=trigger_name
        )
        time.sleep(2)

    def _check_deduced(self, deduced_count, deduced_props, resource_id):
        alarms = self.vitrage_client.alarm.list(vitrage_id=resource_id,
                                                all_tenants=True)
        deduces = g_utils.all_matches(alarms, **deduced_props)
        self.assertThat(deduces, matchers.HasLength(deduced_count),
                        'Expected %s deduces\n - \n%s\n - \n%s' %
                        (deduced_count, alarms, deduces))

    def _check_rca(self, rca, expected_alarms, inspected):
        rca_nodes = [n for n in rca['nodes'] if not n.get('end_timestamp')]
        self.assertEqual(len(expected_alarms), len(rca_nodes))
        for expected_alarm in expected_alarms:
            self.assertIsNotNone(
                g_utils.first_match(rca_nodes, **expected_alarm),
                'expected_alarm is not in the rca %s' % expected_alarm)
        rca_inspected = rca['nodes'][rca['inspected_index']]
        self.assertTrue(g_utils.is_subset(inspected, rca_inspected),
                        'Invalid inspected item \n%s\n%s' %
                        (rca_inspected, inspected))



# Copyright 2017 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import random
import time

from oslo_serialization import jsonutils

from vitrage_tempest_plugin.tests.common.constants import AodhProperties as \
    AodhProps
from vitrage_tempest_plugin.tests.common.tempest_clients import TempestClients


def create_aodh_alarm(resource_id=None, name=None, unic=True):
    if not name:
        name = '%s-%s' % ('test_', random.randrange(0, 100000, 1))
    elif unic:
        name = '%s-%s' % (name, random.randrange(0, 100000, 1))

    aodh_request = _aodh_request('event',
                                 resource_id=resource_id,
                                 name=name)
    TempestClients.aodh().alarm.create(aodh_request)
    time.sleep(20)


def create_aodh_metrics_threshold_alarm(resource_id=None, name=None):
    if not name:
        name = '%s-%s' % ('test_', random.randrange(0, 100000, 1))
    met = TempestClients.gnocchi().metric.create(resource_id=resource_id,
                                                 name='test',
                                                 archive_policy_name='high')
    metric = met[AodhProps.ID]
    rule_opts = dict(
        threshold='100',
        aggregation_method='mean',
        comparison_operator='lt',
        metrics=[metric]
    )
    rule = {
        AodhProps.METRICS_THRESHOLD_RULE: rule_opts
    }
    aodh_request = _aodh_request(AodhProps.METRICS_THRESHOLD,
                                 resource_id=resource_id,
                                 name=name,
                                 rule=rule)
    TempestClients.aodh().alarm.create(aodh_request)
    time.sleep(10)


def create_aodh_resources_threshold_alarm(resource_id, name=None):
    if not name:
        name = '%s-%s' % ('test_', random.randrange(0, 100000, 1))

    gnocchi_resource = _get_gnocchi_resource_id()
    met = TempestClients.gnocchi().metric.create(resource_id=gnocchi_resource,
                                                 name='test',
                                                 archive_policy_name='high')
    metric = met[AodhProps.ID]
    rule_opts = dict(
        threshold='100',
        aggregation_method='mean',
        comparison_operator='lt',
        metric=metric,
        resource_type='nova.instance'
    )
    rule = {
        AodhProps.RESOURCES_THRESHOLD_RULE: rule_opts
    }
    aodh_request = _aodh_request(AodhProps.RESOURCES_THRESHOLD,
                                 resource_id=resource_id,
                                 name=name,
                                 rule=rule)
    TempestClients.aodh().alarm.create(aodh_request)
    time.sleep(10)


def delete_all_aodh_alarms():
    alarms = TempestClients.aodh().alarm.list()
    for alarm in alarms:
        TempestClients.aodh().alarm.delete(alarm[AodhProps.ALARM_ID])
    time.sleep(10)


def delete_all_gnocchi_metrics():
    metrics = TempestClients.gnocchi().metric.list()
    for metric in metrics:
        TempestClients.gnocchi().metric.delete(metric[AodhProps.ID])
    time.sleep(10)


def _get_gnocchi_resource_id():
    return TempestClients.gnocchi().resource.list()[0][AodhProps.ID]


def _aodh_request(type, resource_id=None, name=None, rule=None):
    query = []
    if resource_id:
        query = [
            dict(
                field='traits.resource_id',
                type='',
                op='eq',
                value=resource_id)
        ]
    request = dict(
        name=name,
        description='test alarm',
        severity='low',
        state='alarm',
        type=type)

    if not rule:
        rule = dict(event_rule=(dict(query=query)))
    elif rule.get(AodhProps.RESOURCES_THRESHOLD_RULE):
        # aggregation_by_resources_threshold requests are different from other
        # alarms, update accordingly
        query[0].update(dict(field='resource_id'))
        rule[AodhProps.RESOURCES_THRESHOLD_RULE].update(dict(
            query=jsonutils.dump_as_bytes(query)))

    request.update(rule)
    return request



# Copyright 2016 Nokia
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from testtools import matchers

from vitrage_tempest_plugin.tests.base import BaseVitrageTempest
from vitrage_tempest_plugin.tests.base import IsNotEmpty
from vitrage_tempest_plugin.tests.common import general_utils as g_utils
from vitrage_tempest_plugin.tests.common.general_utils\
    import tempest_resources_dir
from vitrage_tempest_plugin.tests.common import vitrage_utils

LOG = logging.getLogger(__name__)


class BaseTemplateTest(BaseVitrageTempest):
    """Template test class for Vitrage API tests."""

    TEST_PATH = tempest_resources_dir() + '/templates/api/'

    NON_EXIST_FILE = 'non_exist_file.yaml'
    ERROR_FILE = 'corrupted_template.yaml'
    OK_FILE = 'nagios_alarm_for_alarms.yaml'

    VALIDATION_FAILED = 'validation failed'
    VALIDATION_OK = 'validation OK'
    OK_MSG = 'Template validation is OK'

    def tearDown(self):
        super(BaseTemplateTest, self).tearDown()
        self._delete_templates()

    def _delete_templates(self):
        templates = self.vitrage_client.template.list()
        template_ids = [template['uuid'] for template in templates]
        self.vitrage_client.template.delete(template_ids)

    def _compare_template_lists(self, api_templates, cli_templates):
        self.assertThat(api_templates, IsNotEmpty(),
                        'The template list taken from api is empty')
        self.assertIsNotNone(cli_templates,
                             'The template list taken from cli is empty')

        LOG.debug("The template list taken from cli is : %s", cli_templates)
        LOG.debug("The template list taken by api is : %s", api_templates)

        self._validate_templates_list_length(api_templates, cli_templates)
        self._validate_passed_templates_length(api_templates, cli_templates)
        self._compare_each_template_in_list(api_templates, cli_templates)

    def _validate_templates_list_length(self, api_templates, cli_templates):
        self.assertEqual(len(cli_templates.splitlines()),
                         len(api_templates) + 4)

    def _validate_passed_templates_length(self, api_templates, cli_templates):
        api_passes_templates = g_utils.all_matches(
            api_templates,
            **{'status details': self.OK_MSG})
        cli_passes_templates = cli_templates.count(' ' + self.OK_MSG + ' ')
        self.assertThat(api_passes_templates,
                        matchers.HasLength(cli_passes_templates))

    def _compare_each_template_in_list(self, api_templates, cli_templates):
        counter = 0
        for api_item in api_templates:
            for line in cli_templates.splitlines():
                name_start = line.count(' ' + api_item['name'] + ' ')
                status_start = line.count(' ' + api_item['status'] + ' ')
                if name_start > 0 and status_start > 0:
                    counter += 1
                    break
        self.assertThat(api_templates, matchers.HasLength(counter))

    def _assert_validate_result(self, validation, path, negative=False,
                                status_code=0):
        self.assertThat(validation['results'], matchers.HasLength(1))
        result = validation['results'][0]
        self.assertIn(path, result['file path'])

        if negative:
            self.assertEqual(self.VALIDATION_FAILED, result['status'])
            self.assertNotEqual(result['message'], self.OK_MSG)
            self.assertEqual(status_code, result['status code'])
            return

        self.assertEqual(self.VALIDATION_OK, result['status'])
        self.assertEqual(self.OK_MSG, result['message'])
        self.assertEqual(0, result['status code'])

    def _assert_add_result(self, result, status, message):
        self.assertThat(result, matchers.HasLength(1))
        self.assertEqual(status, result[0]['status'])
        self.assertThat(result[0]['status details'],
                        matchers.StartsWith(message))

    @staticmethod
    def _rollback_to_default(templates):
        for t in templates:
            db_row = vitrage_utils.get_first_template(name=t)
            vitrage_utils.delete_template(db_row['uuid'])



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import testtools

from oslo_log import log as logging
from vitrage_tempest_plugin.tests import utils

from vitrage_tempest_plugin.tests.base import BaseVitrageTempest
from vitrage_tempest_plugin.tests.common import aodh_utils
from vitrage_tempest_plugin.tests.common import nova_utils
from vitrage_tempest_plugin.tests.common.tempest_clients import TempestClients

LOG = logging.getLogger(__name__)


class TestAodhAlarm(BaseVitrageTempest):
    NUM_INSTANCE = 1
    NUM_ALARM = 1

    def tearDown(self):
        super(TestAodhAlarm, self).tearDown()
        aodh_utils.delete_all_aodh_alarms()
        aodh_utils.delete_all_gnocchi_metrics()
        nova_utils.delete_all_instances()

    @utils.tempest_logger
    def test_alarm_with_resource_id(self):
        # Action
        nova_utils.create_instances(num_instances=self.NUM_INSTANCE)
        aodh_utils.create_aodh_alarm(self._find_instance_resource_id())
        # Calculate expected results
        api_graph = self.vitrage_client.topology.get(all_tenants=True)
        graph = self._create_graph_from_graph_dictionary(api_graph)
        entities = self._entities_validation_data(
            host_entities=1,
            host_edges=1 + self.NUM_INSTANCE,
            instance_entities=self.NUM_INSTANCE,
            instance_edges=2 * self.NUM_INSTANCE + self.NUM_ALARM,
            aodh_entities=self.NUM_ALARM,
            aodh_edges=self.NUM_ALARM)
        num_entities = self.num_default_entities + \
            2 * self.NUM_INSTANCE + self.NUM_ALARM + \
            self.num_default_networks + self.num_default_ports
        num_edges = self.num_default_edges + 3 * self.NUM_INSTANCE + \
            self.NUM_ALARM + self.num_default_ports

        # Test Assertions
        self._validate_graph_correctness(graph,
                                         num_entities,
                                         num_edges,
                                         entities)

    @utils.tempest_logger
    def test_gnocchi_aggregation_by_metrics_threshold_alarm(self):
        # Action
        nova_utils.create_instances(num_instances=self.NUM_INSTANCE)
        aodh_utils.create_aodh_metrics_threshold_alarm()

        api_graph = self.vitrage_client.topology.get(all_tenants=True)
        graph = self._create_graph_from_graph_dictionary(api_graph)
        entities = self._entities_validation_data(
            host_entities=1,
            host_edges=1 + self.NUM_INSTANCE,
            instance_entities=self.NUM_INSTANCE,
            instance_edges=2 * self.NUM_INSTANCE,
            aodh_entities=self.NUM_ALARM,
            aodh_edges=0)
        num_entities = self.num_default_entities + \
            2 * self.NUM_INSTANCE + self.NUM_ALARM + \
            self.num_default_networks + self.num_default_ports
        num_edges = self.num_default_edges + 3 * self.NUM_INSTANCE + \
            self.num_default_ports

        self._validate_graph_correctness(graph,
                                         num_entities,
                                         num_edges,
                                         entities)

    @testtools.skip('gnocchi sometime fails here ignore for now')
    @utils.tempest_logger
    def test_gnocchi_aggregation_by_resources_threshold(self):
        # Action
        nova_utils.create_instances(num_instances=self.NUM_INSTANCE)
        aodh_utils.create_aodh_resources_threshold_alarm(
            resource_id=self._find_instance_resource_id())

        api_graph = self.vitrage_client.topology.get(all_tenants=True)
        graph = self._create_graph_from_graph_dictionary(api_graph)
        entities = self._entities_validation_data(
            host_entities=1,
            host_edges=1 + self.NUM_INSTANCE,
            instance_entities=self.NUM_INSTANCE,
            instance_edges=2 * self.NUM_INSTANCE + self.NUM_ALARM,
            aodh_entities=self.NUM_ALARM,
            aodh_edges=self.NUM_ALARM)
        num_entities = self.num_default_entities + \
            2 * self.NUM_INSTANCE + self.NUM_ALARM + \
            self.num_default_networks + self.num_default_ports
        num_edges = self.num_default_edges + 3 * self.NUM_INSTANCE + \
            + self.NUM_ALARM + self.num_default_ports

        self._validate_graph_correctness(graph,
                                         num_entities,
                                         num_edges,
                                         entities)

    @utils.tempest_logger
    def test_alarm_without_resource_id(self):
        # Action
        aodh_utils.create_aodh_alarm()

        # Calculate expected results
        api_graph = self.vitrage_client.topology.get(all_tenants=True)
        graph = self._create_graph_from_graph_dictionary(api_graph)
        entities = self._entities_validation_data(
            host_entities=1,
            host_edges=1,
            aodh_entities=self.NUM_ALARM,
            aodh_edges=0)
        num_entities = self.num_default_entities + self.NUM_ALARM + \
            self.num_default_networks + self.num_default_ports
        num_edges = self.num_default_edges + self.num_default_ports

        # Test Assertions
        self._validate_graph_correctness(graph,
                                         num_entities,
                                         num_edges,
                                         entities)

    @staticmethod
    def _find_instance_resource_id():
        servers = TempestClients.nova().servers.list()
        return servers[0].id



# Copyright 2017 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import time
from vitrage_tempest_plugin.tests.common.tempest_clients import TempestClients
from vitrage_tempest_plugin.tests.utils import wait_for_status


def create_volume_and_attach(name, size, instance_id, mount_point):
    volume = TempestClients.cinder().volumes.create(name=name,
                                                    size=size)
    time.sleep(2)
    TempestClients.cinder().volumes.attach(volume=volume,
                                           instance_uuid=instance_id,
                                           mountpoint=mount_point)
    wait_for_status(30, _check_num_volumes, num_volumes=1, state='in-use')
    time.sleep(2)
    return volume


def delete_all_volumes():
    cinder = TempestClients.cinder()
    volumes = cinder.volumes.list()
    for volume in volumes:
        try:
            cinder.volumes.detach(volume)
            cinder.volumes.force_delete(volume)
        except Exception:
            cinder.volumes.reset_state(volume, state='available',
                                       attach_status='detached')
            cinder.volumes.force_delete(volume)
    wait_for_status(30, _check_num_volumes, num_volumes=0)
    time.sleep(2)


def _check_num_volumes(num_volumes=0, state=''):
    if len(TempestClients.cinder().volumes.list()) != num_volumes:
        return False

    return all(vars(volume)['status'].upper() == state.upper() and
               len(vars(volume)['attachments']) == 1
               for volume in TempestClients.cinder().volumes.list())



# Copyright 2018 Nokia
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from datetime import datetime
from oslo_log import log as logging
from testtools import matchers

from vitrage_tempest_plugin.tests.api.event.base import BaseTestEvents
from vitrage_tempest_plugin.tests.base import IsEmpty
from vitrage_tempest_plugin.tests.common.constants import EntityCategory
from vitrage_tempest_plugin.tests.common.constants import VertexProperties \
    as VProps
from vitrage_tempest_plugin.tests.utils import wait_for_answer

LOG = logging.getLogger(__name__)
FIRING = 'firing'
RESOLVED = 'resolved'


class TestPrometheus(BaseTestEvents):
    """Test class for Prometheus datasource"""

    def test_send_prometheus_alert_on_host_without_resource_id(self):
        """Sending an alert in Prometheus format should result in an alarm"""
        self._test_send_prometheus_alert(
            self._create_prometheus_alert_on_host_details('host123', FIRING))

    def test_send_prometheus_alert_on_instance_without_resource_id(self):
        """Sending an alert in Prometheus format should result in an alarm"""
        self._test_send_prometheus_alert(
            self._create_prometheus_alert_on_instance_details('host457',
                                                              'instance457',
                                                              FIRING))

    def _test_send_prometheus_alert(self, details):
        # post an event to the message bus
        event_time = datetime.now()
        event_type = 'prometheus.alarm'
        self._post_event(event_time, event_type, details)
        api_alarms = wait_for_answer(2, 0.5, self._check_alarms)

        # expect to get a 'prometheus.alarm', generated by Prometheus
        # datasource
        self.assertThat(api_alarms, matchers.HasLength(1),
                        'Expected Prometheus alarm')

        alarm = api_alarms[0]
        self._check_alarm(alarm, details)

        event_time = datetime.now()
        self._resolve_alerts(details)
        self._post_event(event_time, event_type, details)

        api_alarms = wait_for_answer(2, 0.5, self._check_alarms)
        self.assertIsNotNone(api_alarms)
        self.assertThat(api_alarms, IsEmpty(), 'Expected no alarms')

    def _check_alarm(self, alarm, details):
        self.assertEqual(EntityCategory.ALARM, alarm[VProps.VITRAGE_CATEGORY])
        self.assertEqual(details['status'], alarm['status'])
        self.assertFalse(alarm[VProps.VITRAGE_IS_DELETED])
        self.assertFalse(alarm[VProps.VITRAGE_IS_PLACEHOLDER])

    @staticmethod
    def _resolve_alerts(details):
        details['status'] = RESOLVED
        for alert in details['alerts']:
            alert['status'] = RESOLVED

    @staticmethod
    def _create_prometheus_alert_on_host_details(hostname, status):
        return {
            "status": status,
            "groupLabels": {
                "alertname": "HighInodeUsage"
            },
            "groupKey": "{}:{alertname=\"HighInodeUsage\"}",
            "commonAnnotations": {
                "mount_point": "/%",
                "description": "\"Consider ssh\"ing into instance \"\n",
                "title": "High number of inode usage",
                "value": "96.81%",
                "device": "/dev/vda1%",
                "runbook": "troubleshooting/filesystem_alerts_inodes.md"
            },
            "alerts": [
                {
                    "status": status,
                    "labels": {
                        "severity": "critical",
                        "fstype": "ext4",
                        "instance": hostname,
                        "job": "node",
                        "alertname": "HighInodeUsage",
                        "device": "/dev/vda1",
                        "mountpoint": "/"
                    },
                    "endsAt": "0001-01-01T00:00:00Z",
                    "generatorURL": "http://devstack-4:9090/graph?g0.htm1",
                    "startsAt": "2018-05-03T12:25:38.231388525Z",
                    "annotations": {
                        "mount_point": "/%",
                        "description": "\"Consider ssh\"ing into instance\"\n",
                        "title": "High number of inode usage",
                        "value": "96.81%",
                        "device": "/dev/vda1%",
                        "runbook": "filesystem_alerts_inodes.md"
                    }
                }
            ],
            "version": "4",
            "receiver": "vitrage",
            "externalURL": "http://devstack-rocky-4:9093",
            "commonLabels": {
                "severity": "critical",
                "fstype": "ext4",
                "instance": hostname,
                "job": "node",
                "alertname": "HighInodeUsage",
                "device": "/dev/vda1",
                "mountpoint": "/"
            }
        }

    @staticmethod
    def _create_prometheus_alert_on_instance_details(hostname,
                                                     instance,
                                                     status):
        return {
            'status': status,
            'groupLabels': {
                'alertname': 'HighCpuOnVmAlert'
            },
            'groupKey': '{}:{alertname="HighCpuOnVmAlert"}',
            'commonAnnotations': {
                'description': 'Test alert to test libvirt exporter.\n',
                'title': 'High cpu usage on vm'
            },
            'alerts': [
                {
                    'status': status,
                    'labels': {
                        'instance': hostname,
                        'domain': instance,
                        'job': 'libvirt',
                        'alertname': 'HighCpuOnVmAlert',
                        'severity': 'critical'
                    },
                    'endsAt': '2019-01-15T12:29:50.91446215Z',
                    'generatorURL': 'http://devstack-rocky-release-4:9090',
                    'startsAt': '2019-01-15T12:26:50.91446215Z',
                    'annotations': {
                        'description': 'Test alert to test libvirt exporter',
                        'title': 'High cpu usage on vm'
                    }
                }
            ],
            'version': '4',
            'receiver': 'vitrage',
            'externalURL': 'http://devstack-rocky-release-4:9093',
            'commonLabels': {
                'instance': hostname,
                'job': 'libvirt',
                'severity': 'critical',
                'alertname': 'HighCpuOnVmAlert'
            }
        }



# Copyright 2015
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

service_available_group = cfg.OptGroup(name="service_available",
                                       title="Available OpenStack Services")

ServiceAvailableGroup = [cfg.BoolOpt("vitrage",
                                     default=True,
                                     help="Whether or not vitrage is expected "
                                     "to be available")]

rca_service_group = cfg.OptGroup(name="root_cause_analysis_service",
                                 title="Root Cause Analysis Service Options")

RcaServiceGroup = [
    # RCA Service tempest configuration
    cfg.IntOpt('instances_per_host',
               default=2,
               help="Number of instances per host in mock graph datasource"),
    cfg.IntOpt('snapshots_interval',
               default=120,
               min=10,
               help='Time to wait between subsequent datasource snapshots'),
    cfg.IntOpt('zabbix_alarms_per_host', default=2),
    cfg.StrOpt('aodh_version', default='2', help='Aodh version'),
    cfg.StrOpt('nova_version', default='2.11', help='Nova version'),
    cfg.StrOpt('cinder_version', default='3', help='Cinder version'),
    cfg.StrOpt('glance_version', default='2', help='Glance version'),
    cfg.StrOpt('heat_version', default='1', help='Heat version'),
    cfg.StrOpt('mistral_version', default='2', help='Mistral version'),
    cfg.StrOpt('gnocchi_version', default='1', help='Gnocchi version'),
    cfg.StrOpt('trove_version', default='1', help='Trove version'),
]



# Copyright 2017 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import ast

import http.server as BaseHTTPServer
from oslo_log import log as logging
import requests
import socket
from testtools import matchers
from threading import Thread
import time

from vitrage_tempest_plugin.tests.base import IsEmpty
from vitrage_tempest_plugin.tests.common.constants import VertexProperties as \
    VProps
from vitrage_tempest_plugin.tests.common import vitrage_utils as v_utils

from vitrage_tempest_plugin.tests.e2e.test_actions_base import TestActionsBase
from vitrage_tempest_plugin.tests import utils


LOG = logging.getLogger(__name__)

TRIGGER_ALARM_1 = 'e2e.test_webhook.alarm1'
TRIGGER_ALARM_2 = 'e2e.test_webhook.alarm2'
TRIGGER_ALARM_WITH_DEDUCED = 'e2e.test_webhook.alarm_with_deduced'
URL = 'url'
REGEX_FILTER = 'regex_filter'
HEADERS = 'headers'
HEADERS_PROPS = '{"content": "application/json"}'
NAME_FILTER = '{"name": "e2e.*"}'
NAME_FILTER_FOR_DEDUCED = '{"name": "e2e.test_webhook.deduced"}'
TYPE_FILTER = '{"vitrage_type": "doctor"}'
FILTER_NO_MATCH = '{"name": "NO MATCH"}'
NOTIFICATION = 'notification'
PAYLOAD = 'payload'
MAIN_FILTER = {NOTIFICATION,
               PAYLOAD}
DOCTOR_ALARM_FILTER = {VProps.VITRAGE_ID,
                       VProps.RESOURCE,
                       VProps.NAME,
                       VProps.UPDATE_TIMESTAMP,
                       VProps.VITRAGE_TYPE,
                       VProps.VITRAGE_CATEGORY,
                       VProps.STATE,
                       VProps.VITRAGE_OPERATIONAL_SEVERITY}
RESOURCE_FILTER = {VProps.VITRAGE_ID,
                   VProps.ID,
                   VProps.NAME,
                   VProps.VITRAGE_CATEGORY,
                   VProps.UPDATE_TIMESTAMP,
                   VProps.VITRAGE_OPERATIONAL_STATE,
                   VProps.VITRAGE_TYPE}
messages = []


class TestWebhook(TestActionsBase):

    @classmethod
    def setUpClass(cls):
        super(TestWebhook, cls).setUpClass()
        cls._template = v_utils.add_template("e2e_webhooks.yaml")
        # Configure mock server.
        cls.mock_server_port = _get_free_port()
        cls.mock_server = MockHTTPServer(('localhost', cls.mock_server_port),
                                         MockServerRequestHandler)

        # Start running mock server in a separate thread.
        cls.mock_server_thread = Thread(target=cls.mock_server.serve_forever)
        cls.mock_server_thread.setDaemon(True)
        cls.mock_server_thread.start()
        cls.URL_PROPS = 'http://localhost:%s/' % cls.mock_server_port

    @classmethod
    def tearDownClass(cls):
        if cls._template is not None:
            v_utils.delete_template(cls._template['uuid'])

    def tearDown(self):
        super(TestWebhook, self).tearDown()
        del messages[:]
        self._delete_webhooks()
        self.mock_server.reset_requests_list()

    @utils.tempest_logger
    def test_basic_event(self):

        try:

            # Add webhook with filter matching alarm
            self.vitrage_client.webhook.add(url=self.URL_PROPS,
                                            regex_filter=NAME_FILTER,
                                            headers=HEADERS_PROPS)

            # Raise alarm
            self._trigger_do_action(TRIGGER_ALARM_1)

            # Check event received
            self.assertThat(self.mock_server.requests, matchers.HasLength(1),
                            'Wrong number of notifications for raise alarm')

            # Undo
            self._trigger_undo_action(TRIGGER_ALARM_1)

            # Check event undo received
            self.assertThat(self.mock_server.requests, matchers.HasLength(2),
                            'Wrong number of notifications for clear alarm')

        finally:
            self._trigger_undo_action(TRIGGER_ALARM_1)

    @utils.tempest_logger
    def test_with_no_filter(self):
        """Test to see that a webhook with no filter receives all

        notifications
        """

        try:

            # Add webhook
            self.vitrage_client.webhook.add(url=self.URL_PROPS,
                                            regex_filter=NAME_FILTER)

            # Raise alarm
            self._trigger_do_action(TRIGGER_ALARM_1)

            # Check event received
            self.assertThat(self.mock_server.requests, matchers.HasLength(1),
                            'Wrong number of notifications for raise alarm')

            # Raise another alarm
            self._trigger_do_action(TRIGGER_ALARM_2)

            # Check second event received
            self.assertThat(self.mock_server.requests, matchers.HasLength(2),
                            'Wrong number of notifications for clear alarm')

        finally:
            self._trigger_undo_action(TRIGGER_ALARM_1)
            self._trigger_undo_action(TRIGGER_ALARM_2)

    @utils.tempest_logger
    def test_with_no_match(self):
        """Test to check that filters with no match do not send event """

        try:

            # Add webhook
            self.vitrage_client.webhook.add(url=self.URL_PROPS,
                                            regex_filter=FILTER_NO_MATCH)

            # Raise alarm
            self._trigger_do_action(TRIGGER_ALARM_1)

            # Check event not received
            self.assertThat(self.mock_server.requests, IsEmpty(),
                            'event should not have passed filter')

            # Raise another alarm
            self._trigger_do_action(TRIGGER_ALARM_2)

            # Check second event not received
            self.assertThat(self.mock_server.requests, IsEmpty(),
                            'event should not have passed filter')

        finally:
            self._trigger_undo_action(TRIGGER_ALARM_1)
            self._trigger_undo_action(TRIGGER_ALARM_2)

    @utils.tempest_logger
    def test_multiple_webhooks(self):
        """Test to check filter by type and with no filter (with 2 separate

        webhooks)
        """

        try:

            # Add webhook
            self.vitrage_client.webhook.add(url=self.URL_PROPS,
                                            regex_filter=TYPE_FILTER)

            self.vitrage_client.webhook.add(url=self.URL_PROPS)

            # Raise alarm
            self._trigger_do_action(TRIGGER_ALARM_1)

            # Check event received
            self.assertThat(self.mock_server.requests, matchers.HasLength(2),
                            'event not posted to all webhooks')

            # Raise another alarm
            self._trigger_do_action(TRIGGER_ALARM_2)

            # Check second event received
            self.assertThat(self.mock_server.requests, matchers.HasLength(4),
                            'event not posted to all webhooks')

        finally:
            self._trigger_undo_action(TRIGGER_ALARM_1)
            self._trigger_undo_action(TRIGGER_ALARM_2)

    @utils.tempest_logger
    def test_for_deduced_alarm(self):

        try:
            # Add webhook with filter for the deduced alarm
            self.vitrage_client.webhook.add(
                url=self.URL_PROPS,
                regex_filter=NAME_FILTER_FOR_DEDUCED,
                headers=HEADERS_PROPS)

            # Raise the trigger alarm
            self._trigger_do_action(TRIGGER_ALARM_WITH_DEDUCED)

            # Check event received - expected one for the deduced alarm
            # (the trigger alarm does not pass the filter). This test verifies
            # that the webhook is called only once for the deduced alarm.
            time.sleep(1)
            self.assertThat(self.mock_server.requests, matchers.HasLength(1),
                            'Wrong number of notifications for deduced alarm')

            # Undo
            self._trigger_undo_action(TRIGGER_ALARM_WITH_DEDUCED)

            # Check event undo received
            time.sleep(1)
            self.assertThat(self.mock_server.requests, matchers.HasLength(2),
                            'Wrong number of notifications '
                            'for clear deduced alarm')

        finally:
            self._trigger_undo_action(TRIGGER_ALARM_WITH_DEDUCED)

    @utils.tempest_logger
    def test_payload_format(self):

        try:

            self.vitrage_client.webhook.add(url=self.URL_PROPS,
                                            headers=HEADERS_PROPS)

            # Raise the trigger alarm
            self._trigger_do_action(TRIGGER_ALARM_1)

            # pre check that correct amount of notifications sent
            self.assertThat(self.mock_server.requests, matchers.HasLength(1),
                            'Wrong number of notifications for alarm')
            self.assertThat(messages, matchers.HasLength(1),
                            'Wrong number of messages for alarm')

            alarm = ast.literal_eval(messages[0])

            # check that only specified fields are sent for the alarm,
            # payload and resource
            passed_filter = utils.filter_data(alarm,
                                              MAIN_FILTER,
                                              match_filter=False)

            self.assertThat(passed_filter,
                            IsEmpty(),
                            "Wrong main fields sent")

            payload = alarm.get(PAYLOAD)
            if payload:
                passed_filter = utils.filter_data(payload,
                                                  DOCTOR_ALARM_FILTER,
                                                  match_filter=False)

                self.assertThat(passed_filter,
                                IsEmpty(),
                                "Wrong alarm fields sent")

                sent_fields = utils.filter_data(payload, DOCTOR_ALARM_FILTER)

                self.assert_set_equal(DOCTOR_ALARM_FILTER, sent_fields,
                                      "Some alarm fields not sent")

                resource = payload.get(VProps.RESOURCE)
                if resource:
                    passed_filter = utils.filter_data(resource,
                                                      RESOURCE_FILTER,
                                                      match_filter=False)

                    self.assertThat(passed_filter,
                                    IsEmpty(),
                                    "Wrong resource fields sent")

                    sent_fields = utils.filter_data(resource, RESOURCE_FILTER)

                    self.assert_set_equal(RESOURCE_FILTER, sent_fields,
                                          "Some resource fields not sent")
        finally:
            self._trigger_undo_action(TRIGGER_ALARM_1)

    @classmethod
    def _delete_webhooks(cls):
        webhooks = cls.vitrage_client.webhook.list()
        for webhook in webhooks:
            cls.vitrage_client.webhook.delete(webhook['id'])


def _get_free_port():

    s = socket.socket(socket.AF_INET, type=socket.SOCK_STREAM)
    s.bind(('localhost', 0))
    address, port = s.getsockname()
    s.close()
    return port


class MockHTTPServer(BaseHTTPServer.HTTPServer):

    def __init__(self, server, handler):
        BaseHTTPServer.HTTPServer.__init__(self, server, handler)
        self.requests = []

    def process_request(self, request, client_address):
        self.requests.append(request)
        LOG.info('received request: %s', request)
        BaseHTTPServer.HTTPServer.process_request(
            self, client_address=client_address, request=request)

    def reset_requests_list(self):
        self.requests = []


class MockServerRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):

    def do_POST(self):
        # Process a HTTP Post request and return status code 200

        content_len = int(self.headers.get('content-length', 0))
        # save received JSON
        messages.append(self.rfile.read(content_len).decode('utf-8'))
        # send fake response
        self.send_response(requests.codes.ok)
        self.end_headers()
        return

    def log_message(self, format, *args):
        pass
